"""
Implements a moiety and a moiety stoichiometry. A moiety
is a functional group in a molecule.
A moiety stoichiometry is a combination of a name and an integer
that indicates the repetitions of the moiety.
A moiety is separated from its stoichiometry by MOIETY_SEPARATOR.

Examples of 

   MOIETY STOICHIOMETRY STRING  MOIETY  STOICHIOMETRY
    A_1                          A      1
    A                            A      1
    A_P                          INVALID
"""

from SBMLLint.common import constants as cn
from SBMLLint.common import util

from collections import namedtuple
import numpy as np
import pandas as pd


NameCount = namedtuple("NameCount", "name count")


NULL_STR = ''
SEP = ","  # Separator between moiety name and stoichiometry


############## CLASSES ##################
class Moiety(object):

  def __init__(self, name, other_moietys=[]):
    """
    :param str name:
    :param list-Moiety other_moieties:
    Ensures unique names within other_moietys
    """
    self.name = name
    if all([name != m.name for m in other_moietys]):
      other_moietys.append(self)

  def __repr__(self):
    return self.name

  def __lt__(self, other):
    """
    Enables sorting a list of Moiety
    """
    return self.name < other.name

  def isEqual(self, other):
    return self.name == other.name


class MoietyStoichiometry(object):
  """A Moiety with its replication count."""

  def __init__(self, moiety, stoichiometry,
      separator=cn.MOIETY_SEPARATOR):
    if isinstance(moiety, Moiety):
      self.moiety = moiety
    else:
      self.moiety = Moiety(str(moiety))
    self.stoichiometry = stoichiometry
    self.name = "%s%s%d" % (self.moiety.name,
        separator, stoichiometry)

  def __repr__(self):
    return self.name

  def __lt__(self, other):
    return str(self) < str(other)

  def isEqual(self, other):
    return self.moiety.isEqual(other.moiety) and  \
        (self.stoichiometry == other.stoichiometry)

  @classmethod
  def getMoietys(cls, moiety_stoichiometrys):
    """
    Extract moieties from MoietyStoichiometrys
    """
    moietys = util.uniqueify([m_s.moiety 
        for m_s in moiety_stoichiometrys])
    return moietys

  @classmethod
  def make(cls, moiety_stoich_stg):
    """
    Makes a MoietyStoichiometry from a string.
    Examples of strings are: "P_2", "A"
    :return MoietyStoichiometry:
    """
    result = []
    pos = moiety_stoich_stg.find(cn.MOIETY_SEPARATOR)
    if pos < 0:
      name = moiety_stoich_stg
      stoich_stg = 1
    elif pos == 0:
      raise ValueError(
          "Invalid format for moiety stoichiometry string: %s"
          % moiety_stoich_stg)
    else:
      name = moiety_stoich_stg[0:pos]
      stoich_stg = moiety_stoich_stg[pos+1:]
    try:
      stoich = int(stoich_stg)
    except ValueError:
      raise ValueError(
          "Invalid number in moiety stoichiometry string: %s"
          % moiety_stoich_stg)
    return cls(Moiety(name), stoich)

  @classmethod
  def makeFromDct(cls, ms_strs):
    """
    Makes a MoietyStoichiometry from the YAML extracted
    for the key cn.CFG_MOIETY_STRUCTURE.
    :param list-str ms_strs: list of moiety stroichiometry strings
    :return list-MoietyStoichiometry:
    """
    result = []
    names = []
    for ms_str in ms_strs:
      terms = [v.strip() for v in ms_str.split(SEP)]
      result.append(MoietyStoichiometry(terms[0], int(terms[1])))
      names.append(terms[0])
    indicies = sorted(range(len(names)), key=lambda k: names[k])
    arr = np.array(result)
    return list(arr[indicies])



"""GAMES Plus (Reduced) Row Echelon Form (GAMES_PP)"""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.common import util
from SBMLLint.games.som import SOM
from SBMLLint.common.simple_sbml import SimpleSBML

import collections
import itertools
import networkx as nx
import numpy as np
import pandas as pd
from scipy.linalg import lu, inv

GAMESErrors = collections.namedtuple("GAMESErrors", 
    "type_one")
ErrorSummary = collections.namedtuple("ErrorSummary",
    "type errors")
TOLERANCE = 0.0001
TYPE_I = "type1"
TYPE_II = "type2"
TYPE_III = "type3"
CANCELING = "canceling"
ECHELON = "echelon"

class SOMStoichiometry(object):

  def __init__(self, som, stoichiometry):
    if not isinstance(som, SOM):
      raise ValueError("First argument must be a SOM.")
    if not util.isFloat(stoichiometry):
      raise ValueError("Second argument must be a float.")
    self.som = som
    self.stoichiometry = stoichiometry
    self.identifier = self.makeId()

  def __repr__(self):
    return self.makeId()

  def makeId(self):
  	return "%s * %2.2f" % (str(self.som), self.stoichiometry)


class SOMReaction(object):

  def __init__(self, reactants, products, label):
    self.reactants = reactants
    self.products = products
    self.label = label
    self.identifier = self.makeId()
    self.category = self.getCategory()

  def __repr__(self):
    return self.makeId()

  def makeId(self):
    """
    Provides a string representation of the som_reaction
    :return str:
    """
    def makeStoichiometryString(som_stoichiometry):
      num = som_stoichiometry.stoichiometry
      if np.isclose(num, 1.0):
        return ''
      else:
        return "%2.2f " % num
    
    def makeTermCollection(som_stoichiometries):
      """
      Formats a set of terms with stoichiometries.
      :param list-MoleculeStoichiometry:
      :return str:
      """
      term_collection = ''
      for s_s in som_stoichiometries:
        term = "%s%s" % (makeStoichiometryString(s_s), str(s_s.som.identifier))
        if len(term_collection) == 0:
          term_collection += term
        else:
          term_collection += " + " + term
      return term_collection
    #
    reactant_collection = makeTermCollection(self.reactants)
    product_collection = makeTermCollection(self.products)
    #
    reaction_str = "%s: %s -> %s" % (self.label,
        reactant_collection, product_collection)
    reaction_str = reaction_str
    return reaction_str

  def getCategory(self):
    """
    Return category of SOMReaction. 
    Return reaction_n_n
    if none of the above applies.
    Each reactant/product (SOMStoichiometry)
    should have a unique SOM wthin the side. 
    :return str:
    """
    num_reactants = len(self.reactants)
    num_products = len(self.products)
    stoichiometry_reactants = [r.stoichiometry for r \
                                  in self.reactants]
    stoichiometry_products = [p.stoichiometry for p \
                             in self.products]
    for reaction_category in cn.REACTION_SUMMARY_CATEGORIES:
      if reaction_category.predicate(num_reactants, num_products, 
                                     stoichiometry_reactants, 
                                     stoichiometry_products):
        return reaction_category.category
    # if none of the above, return reaction_n_n
    return cn.REACTION_n_n


class GAMES_PP(nx.DiGraph):
  """
  Similar to MESGraph, The Message -GAMES++- algorithm creates
  a directed graph of SOMs, and updates the graph using
  the (reduced) row echolon stoichiometry matrix. 
  There are three graphical errors and two matricial errors.

  <Graphical Errors - Type I, II, and III>
  Type I Error occurs when we find inequality between two molecules
  in the same SOM, because each element in a SOM has the same weight.
  Type II Error implies there is cyclism between molecules, such as
  A < B < C < ... < A, which is physically impossible.
  Finally, Type III Errors when it cannot merge two SOMs because
  there is already an existing arc.

  <Matricial Errors - Canceling and Echelon>
  A canceling error occurs when constructing stoichiometry matrix.
  As the algorithm calculates net stoichiometry using SOMs, 
  if a column (a net som_stoichiometry reaction) has only one sign,
  it causes an arror.
  The Echelon error is similar, but it occurs after the matrix is 
  reduced to echelon, or reduced row echelon form. If there is only
  soms with one sign (+ or -), and this causes mass balance error.
  """
  def __init__(self, simple=None):
    """
    :param SimpleSBML simple:
    """
    self.simple = simple
    self.reactions = self._getNonBoundaryReactions(simple)
    self.molecules = self._getNonBoundaryMolecules(simple, self.reactions)
    self.som_stoichiometry_matrix = None
    # reactions before LU decomposition
    self.reactions_lu = []
    # SOMReactinos before LU decomposition
    self.som_reactions_lu = []
    # SOMReactions after LU decomposition
    self.reduced_som_reactions = []
    # RREF SOMReactions after LU -> RREF
    self.rref_som_reactions = []
    # L matrix from LU decomposition
    self.lower = None
    # L^-1 matrix from LU decomposition
    self.lower_inverse  = None
    # P matrix from 'P'LU decomposition
    self.perm_inverse = None
    # permuted stoichiometry matrix
    self.permuted_matrix = None
    # U matrix from LU decomposition
    self.echelon_df = None
    # RREF operation matrix 
    self.rref_operation = None
    # RREF matrix
    self.rref_df = None
    # Components for SOMGraph
    super(GAMES_PP, self).__init__()
    self.soms = self.initializeSOMs(self.molecules)
    # networkx method
    self.add_nodes_from(self.soms)
    self.identifier = self.makeId()
    #
    # List of errors
    # Mass balance error from U matrix (LP decomposition)
    # If not error was detected from U, the error is from RREF
    self.echelon_errors = []
    # Can't add arc
    self.type_one_errors = []
    # Mass balance error from net stoichiometry 
    self.canceling_errors = []
    # SOM cycle ({A} -> {B} -> ... -> {A})
    self.type_two_errors = []
    # Can't merge nodes
    self.type_three_errors = []
    # Cant't add arcs using SOMs (after LU decomposition)
    #self.type_one_som_errors = set()
    # for error return:
    self.error_summary = []
  
  def __repr__(self):
    return self.identifier
  
  def makeId(self):
    """
    Construct an identifier for the graph.
    :return str:
    """
    identifier = ""
    if self.edges:
      for edge in self.edges:
        identifier = identifier + str(edge[0]) + cn.ARC_ARROW + str(edge[1]) + "\n"
    for key, node in enumerate(nx.isolates(self)):
      identifier = identifier + str(node)
      if key < (len(list(nx.isolates(self)))-1):
          identifier = identifier + cn.KINETICS_SEPARATOR
    
    return identifier
  
  def _getNonBoundaryReactions(self, simple):
    """
    Get list of non-boundary reacetions
    :param SimpleSBML simple:
    :return list-Reaction:
    """
    reactions = []
    for reaction in simple.reactions:
      if reaction.category != cn.REACTION_BOUNDARY:
        reactions.append(reaction)
    return reactions  
  
  def _getNonBoundaryMolecules(self, simple, reactions):
    """
    Get list of non-boundary molecules
    :param SimpleSBML simple:
    :return list-Molecule:
    """
    molecules = set()
    for reaction in reactions:
      reactants = {simple.getMolecule(r.molecule.name) for r in reaction.reactants}
      products = {simple.getMolecule(r.molecule.name) for r in reaction.products}
      molecules = molecules.union(reactants)
      molecules = molecules.union(products)
    return list(molecules)
  
  def initializeSOMs(self, molecules):
    """
    Create a list of one-molecule SOMs
    :param list-Molecule molecules:
    :return list-SOM soms:
    """
    soms = []
    for molecule in molecules:
      if molecule.name == cn.EMPTYSET:
        continue
      else:
        soms.append(SOM({molecule}))
    return soms

  def convertReactionToSOMReaction(self, reaction):
    """
    Convert simpleSBML Reaction to SOMReaction
    :param Reaction reaction:
    :return SOMReaction:
    """ 
    reactants = reaction.reactants
    products = reaction.products
    reactant_soms = list({self.getNode(r.molecule) for r in reactants})
    product_soms = list({self.getNode(p.molecule) for p in products})
    
    def getSumStoichiometry(som, species):
      """
      :param SOM som:
      :param list-MoleculeStoichiometry species:
      :return SOMStoichiometry:
      """   
      sum_stoichiometry = 0.0
      for s in species:
        if self.getNode(s.molecule) == som:
          sum_stoichiometry += s.stoichiometry
      return SOMStoichiometry(som, sum_stoichiometry)
    #
    ss_reactants = []
    ss_products = []
    for som in reactant_soms:
      ss_reactants.append(getSumStoichiometry(som, reactants))
    for som in product_soms:
      ss_products.append(getSumStoichiometry(som, products))
    #
    return SOMReaction(ss_reactants, ss_products, reaction.label)
  
  def getStoichiometryMatrix(self, reactions, species, som=False):
    """
    Creates a full stoichiometry matrix
    using non-boundary reactions.
    Due to the issue with multiple Molecules with same name, 
    species can also be a list of str - in this case, 
    each component must be molecule.name. 
    Helped by https://gist.github.com/lukauskas/d1e30bdccc5b801d341d
    :param list-Reaction/SOMReaction reactions:
    :param list-str/Molecule/SOM species:
    :return pd.DataFrame:
    """
    reaction_labels = [r.label for r in reactions]
    if som:
      species_names = [s.identifier for s in species]
    elif type(species[0]) == str:
      species_names = species
    else:
      species_names = [m.name for m in species]
    stoichiometry_matrix = pd.DataFrame(
        0.0,
        index=species_names,
        columns=reaction_labels
        )
    for reaction in reactions:
      if som:
        reactants = {r.som.identifier:r.stoichiometry for r in reaction.reactants}
        products = {p.som.identifier:p.stoichiometry for p in reaction.products}
      else:
        reactants = {r.molecule.name:r.stoichiometry for r in reaction.reactants}
        products = {p.molecule.name:p.stoichiometry for p in reaction.products}
      reaction_species = list(set(reactants.keys()).union(products.keys()))
      for species_name in reaction_species:
        net_stoichiometry = products.get(species_name, 0.0) - reactants.get(species_name, 0.0)
        stoichiometry_matrix.loc[species_name, reaction.label] = net_stoichiometry
    return stoichiometry_matrix
  
  def decomposeMatrix(self, mat_df):
    """
    LU decomposition of the stoichiometry matrix.
    First it transposes the input matrix
    and find L, U matrices and permutation list. 
    :param pandas.DataFrame mat_df:
    :return pandas.DataFrame echelon_df:
    """
    mat_t = mat_df.T
    idx_mat_t = mat_t.index
    cols_mat_t = mat_t.columns
    #
    diff = None
    if mat_t.shape[0] > mat_t.shape[1]:
      diff = mat_t.shape[0] - mat_t.shape[1]
      for i in range(diff):
        mat_t["_" + str(i)] = np.zeros(mat_t.shape[0])
    # LU decomposition
    perm, lower, upper = lu(mat_t)
    # #### Trying to round up lower and upper, to avoid precision issue related to 0.0
    # lower = np.round(lower_raw, 3)
    # upper = np.round(upper_raw, 3)
    perm_inverse = perm.T
    permuted_m = (perm_inverse).dot(mat_t)
    pivot_index = [list(k).index(1) for k in perm_inverse]
    new_idx_mat_t = [idx_mat_t[idx] for idx in pivot_index]
    # we save as; lower_inverse * perm_df = echelon_df
    # if we added zero columns previously, delete them. 
    if diff:
      perm_df = pd.DataFrame(permuted_m,
          index=new_idx_mat_t,
          columns=mat_t.columns).drop(columns = mat_t.columns[-diff:]).T
      echelon_df = pd.DataFrame(upper,
          index=new_idx_mat_t,
          columns=mat_t.columns).drop(columns = mat_t.columns[-diff:]).T
    else:
      perm_df = pd.DataFrame(permuted_m,
          index=new_idx_mat_t,
          columns=mat_t.columns).T
      echelon_df = pd.DataFrame(upper,
          index=new_idx_mat_t,
          columns=mat_t.columns).T
    lower_inverse = pd.DataFrame(inv(lower),
          index=new_idx_mat_t,
          columns=new_idx_mat_t)
    self.perm_inverse = perm_inverse
    self.permuted_matrix = perm_df
    self.lower = lower
    self.lower_inverse = lower_inverse
    self.echelon_df = echelon_df
    return echelon_df
  
  def getRREFMatrix(self, echelon_df):
    """
    Get RREF of the stoichiometry matrix.
    Stoichiometry matrix should be in echelon form.
    Columns are reactions and indices are species. 
    :param pandas.DataFrame echelon_df:
    :return pandas.DataFrame rref_df:
    """
    rref_operation = np.identity(echelon_df.T.shape[0])
    rref_operation = pd.DataFrame(rref_operation,
                     index = echelon_df.columns,
                     columns = echelon_df.columns)

    # now update the operation matrix, finally multiply (dot product) two matrices
    for idx, colname in enumerate(echelon_df.columns):
      reaction_series = echelon_df[colname]
      # Find the first nonzero values
      # Deprecation: instead of np.nonzero(Series), use Series.to_numpy().nonzero()
      # to fix error on GitHub - not use to_numpy here
      ##nonzero_idx = echelon_df[colname].to_numpy().nonzero()[0]
      nonzero_idx = np.array([idx for idx, val in enumerate(echelon_df[colname]) if val != 0])
      # Skip if there is no nonzero value or if it is first reaction
      if not nonzero_idx.any() or idx == 0:
        continue
      nonzero_species = reaction_series.index[nonzero_idx[0]]
      nonzero_value = reaction_series[nonzero_idx[0]]
      # find current nonzero index
      current_echelon_t = rref_operation.dot(echelon_df.T).T
      for prev_colname in current_echelon_t.columns[:idx]:
        if np.round(current_echelon_t[prev_colname][nonzero_species], 3) != 0.0:
          reduction_value = current_echelon_t[prev_colname][nonzero_species]
          rref_operation.at[prev_colname, colname] = (-1.0) * reduction_value / nonzero_value
    rref_df = np.round(rref_operation.dot(echelon_df.T).T, 3)
    self.rref_operation = rref_operation	
    self.rref_df = rref_df
    return rref_df

  def convertMatrixToSOMReactions(self, mat_df):
    """
    Convert a stoichiometry matrix to SOMReactions,
    where columns are SOMReactions and 
    rows are SOMs (species).
    :param pandas.DataFrame mat_df:
    :return list-SOMReaction reactions:
    """
    reactions = []
    for reaction_name in mat_df.columns:
      reaction_elements = mat_df[reaction_name]
      reactants = [SOMStoichiometry(
          self.getNode(som_label),
          np.round(abs(reaction_elements[som_label]), 3)
          ) \
          for som_label in reaction_elements.index \
          if reaction_elements[som_label]<TOLERANCE*(-1)]
      products = [SOMStoichiometry(
          self.getNode(som_label),
          np.round(abs(reaction_elements[som_label]), 3)
          ) \
          for som_label in reaction_elements.index \
          if reaction_elements[som_label]>TOLERANCE]
      reactions.append(SOMReaction(
          reactants=reactants,
          products=products,
          label=reaction_name
          ))
    return reactions
  
  def getNode(self, input_arg):
    """
    Find a node(SOM) containing the given molecule
    or the SOM identifier. 
    If no such SOM exists, return False
    :param Molecule/str input_arg:
    :return SOM/False:
    """
    for som in list(self.nodes):
      if isinstance(input_arg, Molecule):
        for mole in som.molecules:
          if mole.name == input_arg.name:
            return som
      elif isinstance(input_arg, str):
      	if som.identifier == input_arg:
      	  return som
    return False
  #
  def mergeNodes(self, som1, som2, reaction):
    """
    Merge two nodes (SOMs).
    Update arcs if applicable. 
    :param SOM som1:
    :param SOM som2:
    :param Reaction reaction:
    :return SOM new_som:
    """
    new_som = som1.merge(som2)
    new_som.reactions.add(self.simple.getReaction(reaction.label))
    for som in [som1, som2]:
      for edge in list(self.in_edges(som)):
        remaining_som = edge[0]
        reaction_label = self.get_edge_data(edge[0], edge[1])[cn.REACTION]
        self.add_edge(remaining_som, new_som, reaction=reaction_label)  
      for edge in list(self.out_edges(som)):
        remaining_som = edge[1]
        reaction_label = self.get_edge_data(edge[0], edge[1])[cn.REACTION]
        self.add_edge(new_som, remaining_som, reaction=reaction_label) 
    self.remove_nodes_from([som1, som2])
    if not self.has_node(new_som):
      self.add_node(new_som)
    return new_som
  
  def addReaction(self, reaction=None):
    """
    Add a reaction to self.reactions_lu
    :param reaction Reaction:
    """
    if reaction not in self.reactions_lu:
      self.reactions_lu.append(reaction)  
  
  def processUniUniReaction(self, reaction):
    """
    Process a 1-1 reaction to merge nodes.
    If no need to merge, return None.
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_1_1:
      pass
    else:
      reactant_som = self.getNode(reaction.reactants[0].molecule)
      product_som = self.getNode(reaction.products[0].molecule)
      if reactant_som == product_som:
        return None
      else:
        new_som = self.mergeNodes(reactant_som, product_som, reaction)
        self.identifier = self.makeId()
        return new_som
  
  def addArc(self, arc_source, arc_destination, reaction):
    """
    Add a single arc (edge) using two SOMs and reaction/somreaction.
    :param SOM arc_source:
    :param SOM arc_destination:
    :param Reaction/SOMReaction reaction:
    """
    # if there is already a preious reaction,
    if self.has_edge(arc_source, arc_destination):
      reaction_label = self.get_edge_data(arc_source, arc_destination)[cn.REACTION]
      # if reaction.label is not already included in the attribute,
      if reaction.label not in set(reaction_label):
        reaction_label = reaction_label + [reaction.label]
    else:
      reaction_label = [reaction.label]
    # overwrite the edge with new reactions set
    self.add_edge(arc_source, arc_destination, reaction=reaction_label)
  
  def processUniMultiReaction(self, reaction):
    """
    Process a 1-n reaction to add arcs.
    Since the mass of reactant is greater than
    that of each product, it adds arcs by
    addArc(source=products, destination=reactant).
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_1_n:
      pass
    else:
      destination = [reaction.reactants[0].molecule]
      source = [product.molecule for product in reaction.products]
      arcs = itertools.product(source, destination)
      error_count = 0
      for arc in arcs:
        if not self.checkTypeOneError(arc, reaction):
          som_source = self.getNode(arc[0])
          som_destination = self.getNode(arc[1])
          self.addArc(som_source, som_destination, reaction)
        else:
          error_count += 1
      # the reaction has not added any errors, will be used for lu
      if error_count == 0:
        self.reactions_lu.append(reaction)
      self.identifier = self.makeId()
  
  def processMultiUniReaction(self, reaction):
    """
    Process a n-1 reaction to add arcs.
    Since the mass of product is greater than
    that of each reactant, it adds arcs by
    addArc(source=reactants, destination=product).
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_n_1:
      pass
    else:
      destination = [reaction.products[0].molecule]
      source = [reactant.molecule for reactant in reaction.reactants]
      arcs = itertools.product(source, destination)
      error_count = 0
      for arc in arcs:
        if not self.checkTypeOneError(arc, reaction):
          som_source = self.getNode(arc[0])
          som_destination = self.getNode(arc[1])
          self.addArc(som_source, som_destination, reaction)
        else:
          error_count += 1
      # the reaction has not added any errors, will be used for lu
      if error_count == 0:
        self.reactions_lu.append(reaction)
      self.identifier = self.makeId()
  
  def processEqualSOMReaction(self, reaction):
    """
    Process a 1-1 SOMReaction to check 
    mergeability of nodes.
    If there are existing arcs, two nodes
    cannot be merged which will lead to a 
    type_three_error.
    :param SOMReaction reaction:
    """
    if reaction.category != cn.REACTION_1_1:
      pass
    else:
      reactant = reaction.reactants[0].som
      product = reaction.products[0].som
      if self.has_edge(reactant, product) or self.has_edge(product, reactant):
        self.type_three_errors.append(reaction)

  def processUnequalSOMReaction(self, reaction):
    """
    Process a 1-n or n-1 SOMReaction to add arcs.
    An arc flows from SOM with less mass (source)
    to SOM with greater mass (destination). 
    :param SOMReaction reaction:
    """
    category = reaction.category
    if (category!=cn.REACTION_n_1) and (category!=cn.REACTION_1_n):
      pass
    else:
      if category == cn.REACTION_n_1:
        destination = [reaction.products[0].som]
        source = [reactant.som for reactant in reaction.reactants]
      elif category == cn.REACTION_1_n:
        destination = [reaction.reactants[0].som]
        source = [product.som for product in reaction.products]
      arcs = itertools.product(source, destination)
      for arc in arcs:
        if arc[0] == arc[1]:
          #self.type_one_som_errors = self.type_one_som_errors.add(reaction.label)
          raise ValueError("SOM Unequal reaction leads an arc between two soms")
        else:
          self.addArc(arc[0], arc[1], reaction)
      self.identifier = self.makeId()
  
  def addTypeOneError(self, mole1, mole2, reaction):
    """
    Add Type I Error components to self.type_one_errors
    All components of resulting PathComponents are str
    :param Molecule mole1:
    :param Molecule mole2:
    :param Reaction reaction:
    :return bool flag:
    """
    flag = False
    for component in self.type_one_errors:
      if (component.node1==mole1.name) and (component.node2==mole2.name):
        new_component = cn.PathComponents(node1=mole1.name, 
                                          node2=mole2.name,
                                          reactions=component.reactions+[reaction.label])
        self.type_one_errors.remove(component)
        self.type_one_errors.append(new_component)
        flag = True
        break
    if not flag:
      self.type_one_errors.append(cn.PathComponents(node1=mole1.name, 
                                                    node2=mole2.name,
                                                    reactions=[reaction.label]))
      flag = True
    return flag
  
  def checkTypeOneError(self, arc, inequality_reaction):
    """
    Check Type I Error of an arc.
    If both source and destination are found
    in the same SOM, send error message and return True.
    If not, return False.
    :param tuple-Molecule arc:
    :param Reaction inequality_reaction:
    :return bool:
    """
    som1 = self.getNode(arc[0])
    som2 = self.getNode(arc[1])
    if som1 == som2:
      self.addTypeOneError(arc[0], arc[1], inequality_reaction)
      return True
    else:
      return False
  
  # We may remove this method
  def addTypeTwoError(self, cycle):
    """
    Add Type II Error components to self.type_two_errors
    which is a list of lists
    All components of resulting PathComponents are str
    :param list-SOM cycle:
    """
    # exceptionally, here PathComponents are
    # node1=[], node2=[], reactions=[] and their index
    # of each component will match. All elements within nodes
    # are in the same SOM
    error_cycle = []
    for node_idx in range(len(cycle)-1):
      som1 = cycle[node_idx]
      som2 = cycle[node_idx+1]
      som1_moles = {mole.name for mole in list(som1.molecules)}
      som2_moles = {mole.name for mole in list(som2.molecules)}
      reactions = self.get_edge_data(som1, som2)[cn.REACTION]
      # all reactions (in an edge), should create a single PathComponent
      nodes1 = []
      nodes2 = []
      reaction_labels = []
      for r in reactions:
        reaction = self.simple.getReaction(r)
        if reaction.category == cn.REACTION_n_1:
          sources = {r.molecule.name for r in reaction.reactants}
          destinations = {p.molecule.name for p in reaction.products}
        elif reaction.category == cn.REACTION_1_n:
          sources = {p.molecule.name for p in reaction.products}
          destinations = {r.molecule.name for r in reaction.reactants}
        # for any reaction that addes arcs, len(nodes2)==1
        node2 = list(destinations.intersection(som2_moles))[0]
        for node1 in list(sources.intersection(som1_moles)):
          nodes1.append(node1)
          nodes2.append(node2)
          reaction_labels.append(reaction.label)
      error_cycle.append(cn.PathComponents(node1=nodes1, 
                                           node2=nodes2,
                                           reactions=reaction_labels))
    som1 = cycle[-1]
    som2 = cycle[0]
    som1_moles = {mole.name for mole in list(som1.molecules)}
    som2_moles = {mole.name for mole in list(som2.molecules)}
    reactions = self.get_edge_data(som1, som2)[cn.REACTION]
    # all reactions (in an edge), should create a single PathComponent
    nodes1 = []
    nodes2 = []
    reaction_labels = []
    for r in reactions:
      reaction = self.simple.getReaction(r)
      if reaction.category == cn.REACTION_n_1:
        sources = {r.molecule.name for r in reaction.reactants}
        destinations = {p.molecule.name for p in reaction.products}
      elif reaction.category == cn.REACTION_1_n:
        sources = {p.molecule.name for p in reaction.products}
        destinations = {r.molecule.name for r in reaction.reactants}
        # for any reaction that addes arcs, len(nodes2)==1
      node2 = list(destinations.intersection(som2_moles))[0]
      for node1 in list(sources.intersection(som1_moles)):
        nodes1.append(node1)
        nodes2.append(node2)
        reaction_labels.append(reaction.label)
    error_cycle.append(cn.PathComponents(node1=nodes1, 
                                        node2=nodes2,
                                        reactions=reaction_labels))
    self.type_two_errors.append(error_cycle)  
  
  def checkTypeTwoError(self):
    """
    Check Type II Error (cycles) of a MESGraph.
    If there is at least one cycle, 
    report an error message, related reactions
    and return True.
    If there is no cycle, return False. 
    :return bool:
    """
    graph = nx.DiGraph()
    graph.add_edges_from(self.edges)
    cycles = list(nx.simple_cycles(graph))
    if len(cycles) == 0:
      return False
    else:
      for cycle in cycles:
        # We are not using addTypeTwoError yet. 
        # self.addTypeTwoError(cycle)
        self.type_two_errors.append(cycle)
        # if not self.type_two_error:
        #   self.type_two_error = True
      return True
  
  def processErrorReaction(self, reaction):
    """
    Simply add error reactions to error
    :param Reaction reaction:
    :return bool:
    """
    error_reactions = [r.label for r in self.echelon_errors]
    if reaction.label not in error_reactions:
      self.echelon_errors.append(reaction)
    ## Here, we may need to add more info that will 
    ## help us track the operations that lead to this error 
    return True
  
  def analyze(self, reactions=None, simple_games=False, rref=True, error_details=False, suppress_message=False):
    """
    Using the stoichiometry matrix, compute
    row reduced echelon form and create SOMGraph
    Add arcs or sending error messages using
    checkTypeOneError or checkTypeTwoError.
    :param list-Reaction reactions:
    :param bool rref:
    :param bool error_details:
    :param bool suppress_message:
    :return bool:
    """
    multimulti_error_found = False
    if reactions is None:
      reactions = self.simple.reactions
    # Associate the reaction category with the function
    # that processes that category
    report = cn.NULL_STR
    reaction_dic = {
        cn.REACTION_1_1: self.processUniUniReaction,
        cn.REACTION_1_n: self.processUniMultiReaction,
        cn.REACTION_n_1: self.processMultiUniReaction,
        cn.REACTION_n_n: self.addReaction,
        }
    # Process each type of reaction - Type I error will be detected here
    for category in reaction_dic.keys():
      for reaction in [r for r in reactions if r.category == category]:
        func = reaction_dic[category]
        func(reaction)
    # detect type II error
    self.checkTypeTwoError()
    #########################
    # if we find type I or II errors, we make it a simple_game
    if self.type_one_errors or self.type_two_errors:
      simple_games = True
    ########################
    # if simple_games, we only run elementary operations
    if not simple_games:
      # reaction_lu are reactions for LU decomposition, i.e. multi-multi reactions
      if self.reactions_lu:
        for reaction in self.reactions_lu:
          self.som_reactions_lu.append(
              self.convertReactionToSOMReaction(reaction)
              )
        # Now, step 1: creates SOMStoichiometryMatrix
        self.som_stoichiometry_matrix = self.getStoichiometryMatrix(self.som_reactions_lu, list(self.nodes), som=True)
        # step 2: reconvert it into SOMReactions and examine 'canceling errors'
        initial_reduced_som_reactions = self.convertMatrixToSOMReactions(self.som_stoichiometry_matrix)
        dropping_reactions = []
        for r in initial_reduced_som_reactions:
          if r.category == cn.REACTION_ERROR:
            self.canceling_errors.append(r)
            dropping_reactions.append(r.label)
            multimulti_error_found = True
        # step 3: decompose using LU decompositon and check errors (echelon, type_three)
        if not multimulti_error_found:
          echelon_df = self.decomposeMatrix(self.som_stoichiometry_matrix)
          self.reduced_som_reactions = self.convertMatrixToSOMReactions(echelon_df)
          som_reaction_dic = {
              cn.REACTION_ERROR: self.processErrorReaction,
              cn.REACTION_1_1: self.processEqualSOMReaction,
              cn.REACTION_1_n: self.processUnequalSOMReaction,
              cn.REACTION_n_1: self.processUnequalSOMReaction,
              }
          for category in som_reaction_dic.keys():
            for reaction in [r for r in self.reduced_som_reactions if r.category == category]:
              func = som_reaction_dic[category]
              func(reaction)
          # checking if there was any error by LU decompoistion
          if self.echelon_errors or self.type_three_errors:
            multimulti_error_found = True
        # step 4: get RREF and check errors (same as LU decomposition case)
        if not multimulti_error_found:
          rref_df = self.getRREFMatrix(self.echelon_df)
          self.rref_som_reactions = self.convertMatrixToSOMReactions(rref_df)
          for category in som_reaction_dic.keys():
            for reaction in [r for r in self.rref_som_reactions if r.category == category]:
              func = som_reaction_dic[category]
              func(reaction)
    if not suppress_message:
      print("Model analyzed...")
    if error_details:
      print("Type I error: ", self.type_one_errors)
      print("Type II error: " , self.type_two_errors)
      print("Canceling error: ", self.canceling_errors)
      print("Echelon error: ", self.echelon_errors)
      print("Type III error: ", self.type_three_errors, "\n")
    if self.echelon_errors or self.type_one_errors or self.type_two_errors \
        or self.canceling_errors or self.type_three_errors:
      if self.type_one_errors:
        self.error_summary.append(ErrorSummary(type=TYPE_I, errors=self.type_one_errors))
      if self.type_two_errors:
        self.error_summary.append(ErrorSummary(type=TYPE_II, errors=self.type_two_errors))
      if self.type_three_errors:
        self.error_summary.append(ErrorSummary(type=TYPE_III, errors=self.type_three_errors))
      if self.canceling_errors:
        self.error_summary.append(ErrorSummary(type=CANCELING, errors=self.canceling_errors))
      if self.echelon_errors:
        self.error_summary.append(ErrorSummary(type=ECHELON, errors=self.echelon_errors))
      if not suppress_message:
        print("At least one error found.\n")
      return True
    else:
      if not suppress_message:
        print("No error found.")
      return False










#!/usr/bin/env python
"""Prints the reactions in a model."""

from SBMLLint.common import constants as cn
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common import util

import argparse
import sys
import libsbml


def prettyPrint(model_reference, file_out=sys.stdout, **kwargs):
  """
  Prints the reactions in a model.
  :param str model_reference: file, xml string, antimony string
  :param dict kwargs: arguments to Reaction.getId
  """
  xml = util.getXML(model_reference)
  reader = libsbml.SBMLReader()
  document = reader.readSBMLFromString(xml)
  util.checkSBMLDocument(document)
  model = document.getModel()
  simple = SimpleSBML()
  simple.initialize(model)
  stgs = []
  for reaction in simple.reactions:
    stg = reaction.getId(**kwargs)
    stgs.append(stg)
    file_out.write("%s\n" % stg)
  return stgs

def main():
  def str2Bool(stg):
    if "T" in stg.upper():
      return True
    if "F" in stg.upper():
      return False
    return False
  #
  parser = argparse.ArgumentParser(description='SBML XML file.')
  parser.add_argument('filename', type=open, help='SBML file')
  parser.add_argument('--kinetics', nargs=1, type=str2Bool,
      help="Print kinetics formula True or False",
      default = ['True'])
  args = parser.parse_args()
  for fid in util.getNextFid(args.filename):
    try:
      prettyPrint(fid, is_include_kinetics=args.kinetics[0])
    except ValueError:
      print ("  *** Bad SBML file.")


if __name__ == '__main__':
  main()



"""
Test for Games Report
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.games_pp import SOMStoichiometry, SOMReaction, GAMES_PP
from SBMLLint.games.games_report import GAMESReport, SimplifiedReaction, NULL_STR, NUM_STAR, PARAGRAPH_DIVIDER, REPORT_DIVIDER
from SBMLLint.games.som import SOM
from SBMLLint.common import simple_sbml

import numpy as np
import pandas as pd
import os
import re
import unittest

IGNORE_TEST = False
# For SimplifiedReaction
CREATINEKINASE = "CreatineKinase"
PCR = "PCr"
CR = "Cr"
# For Type I Error
CDC2PHOS = "Cdc2Phos"
RUM1DEGINPG2R = "Rum1DegInPG2R"
G2R_CREATION = "G2R_Creation"
G2K = "G2K"
G2R = "G2R"
PG2 = "PG2"
PG2R = "PG2R"
# For Type II Error
MTR = "MTR"
MTHFR = "MTHFR"
CH3FH4 = "CH3FH4"
FH4 = "FH4"
# For Echelon, Type III Error
STATPHOSPHORYLATION = "statPhosphorylation"
PSTATDIMERISATION = "PstatDimerisation"
PSTATDIMERISATIONNUC = "PstatDimerisationNuc"
SPECIES_TEST = "species_test"
PSTAT_SOL = "Pstat_sol"
PSTATDIMER_NUC = "PstatDimer_nuc"
PSTAT_NUC = "Pstat_nuc"
STAT_SOL = "stat_sol"
PSTATDIMER_SOL = "PstatDimer_sol"


#############################
# Tests
#############################
class TestSimplifiedReaction(unittest.TestCase):

  def setUp(self):
  	self.simple = SimpleSBML()
  	# BIOMD0000000248 - originally for canceling_error
  	self.simple.initialize(cn.TEST_FILE_GAMESREPORT1)
  	self.mesgraph = GAMES_PP(self.simple)
  	self.mesgraph.analyze(error_details=False)
  	# Construct SimplifiedReaction
  	self.reaction = self.simple.getReaction(CREATINEKINASE)
  	self.simplified_reaction = SimplifiedReaction(self.reaction.reactants,
  		                                          self.reaction.products,
  		                                          self.reaction.label,
  		                                          self.mesgraph)

  def testConstructor(self):
  	if IGNORE_TEST:
  	  return
  	self.assertEqual(type(self.simplified_reaction.reactants[0]), MoleculeStoichiometry)
  	self.assertEqual(type(self.simplified_reaction.products[0]), MoleculeStoichiometry)
  	self.assertEqual(self.simplified_reaction.label, CREATINEKINASE)
  	self.assertEqual(type(self.simplified_reaction.mesgraph), GAMES_PP)
  	self.assertEqual(self.simplified_reaction.makeIdentifier(),
  		             self.reaction.makeIdentifier(is_include_kinetics=False))

  def testReduceBySOMs(self):
  	if IGNORE_TEST:
  	  return
  	self.simplified_reaction.reduceBySOMs()
  	self.assertEqual(len(self.simplified_reaction.reactants), 1)
  	self.assertEqual(len(self.simplified_reaction.products), 1)
  	self.assertTrue(self.simplified_reaction.reactants[0].molecule.name == PCR)
  	self.assertTrue(self.simplified_reaction.products[0].molecule.name == CR)


class TestGAMESReport(unittest.TestCase):

  def setUp(self):
    self.simple1 = SimpleSBML()
    self.simple2 = SimpleSBML()
    self.simple3 = SimpleSBML()
    self.simple4 = SimpleSBML()
    # BIOMD0000000248 - canceling_error
    self.simple1.initialize(cn.TEST_FILE_GAMESREPORT1)
    # BIOMD0000000007 - Type I error
    self.simple2.initialize(cn.TEST_FILE_GAMESREPORT2)
    # BIOMD0000000018 - Type II error
    self.simple3.initialize(cn.TEST_FILE_GAMES_PP2)
    # BIOMD0000000167 - Echelon, Type III error
    self.simple4.initialize(cn.TEST_FILE_GAMESREPORT3)

  def testReportCancelingError(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple1)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    report, error_num = gr.reportCancelingError(m.canceling_errors, explain_details=True)
    extended_report = NULL_STR
    extended_report = extended_report + "We detected a mass imbalance\n"
    extended_report = extended_report + ": OxidativePhosphorylation: CTtis -> \n\n"
    extended_report = extended_report + "from the following isolation set:\n\n"
    extended_report = extended_report + "1. OxidativePhosphorylation: 6.00 ADP + CTtis -> 6.00 ATP\n"
    extended_report = extended_report + "2. ATPase: ATP -> ADP\n"
    extended_report = extended_report + "*ATP and ADP have the same mass according to the above reaction\n"
    extended_report = extended_report + "\n%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
    extended_report = extended_report + "\n%s\n" % REPORT_DIVIDER
    self.assertEqual(extended_report, report)
    self.assertEqual(error_num, [2])

  def testGetMoleculeEqualityPath(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple2)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    som = m.getNode(self.simple2.getMolecule(G2K))
    equality_path = gr.getMoleculeEqualityPath(som, G2K, PG2R)
    self.assertTrue(len(equality_path) == 2)
    self.assertEqual(type(equality_path[0]), cn.PathComponents)
    self.assertEqual(equality_path[0].node1, G2K)
    self.assertEqual(equality_path[0].reactions, [CDC2PHOS])
    self.assertEqual(equality_path[1].node2, PG2R)
    self.assertEqual(equality_path[1].reactions, [RUM1DEGINPG2R]) 

  def testGetMoleculeEqualityPathReport(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple2)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    count, report1 = gr.getMoleculeEqualityPathReport(G2K, PG2R, 0, explain_details=False)
    self.assertEqual(count, 2)
    self.assertEqual(report1,
    	             "1. Cdc2Phos: G2K -> PG2\n2. Rum1DegInPG2R: PG2R -> PG2\n")
    count, report2 = gr.getMoleculeEqualityPathReport(G2K, PG2R, 0, explain_details=True)
    self.assertEqual(count, 2)
    self.assertEqual(report2,
    	             "\nG2K = PG2 by reaction(s):\n1. Cdc2Phos: G2K -> PG2\n\nPG2 = PG2R by reaction(s):\n2. Rum1DegInPG2R: PG2R -> PG2\n")

  def testGetMoleculeInequalityPathReport(self):
  	if IGNORE_TEST:
  	  return
  	m = GAMES_PP(self.simple2)
  	m.analyze(error_details=False)
  	gr = GAMESReport(m)
  	count, report1 = gr.getMoleculeInequalityPathReport(G2K, PG2R, ["G2R_Creation"], 0, explain_details=False)
  	self.assertEqual(count, 1)
  	self.assertEqual(report1,
  		             "1. G2R_Creation: G2K + R -> G2R\n")
  	count, report2 = gr.getMoleculeInequalityPathReport(G2K, PG2R, ["G2R_Creation"], 0, explain_details=True)
  	self.assertEqual(count, 1)
  	self.assertEqual(report2,
  		             "G2K < PG2R by reaction(s):\n1. G2R_Creation: G2K + R -> G2R\n")

  def testReportTypeOneError(self):
  	if IGNORE_TEST:
  	  return
  	m = GAMES_PP(self.simple2)
  	m.analyze(error_details=False)
  	gr = GAMESReport(m)
  	error = [cn.PathComponents(node1=G2K, node2=G2R, reactions=[G2R_CREATION])]
  	report, error_num = gr.reportTypeOneError(error, explain_details=True)
  	self.assertEqual(error_num, [2])
  	extended_report = NULL_STR
  	extended_report = extended_report + "\nG2K = G2R by reaction(s):\n1. Rum1DegInG2R: G2R -> G2K\n\n"
  	extended_report = extended_report + "However, G2K < G2R by reaction(s):\n2. G2R_Creation: G2K + R -> G2R\n\n"
  	extended_report = extended_report + "\n----------------------------------------------------------------------\n\n"
  	extended_report = extended_report + "\n\n**********************************************************************\n\n"
  	self.assertEqual(report, extended_report)

  def testReportTypeTwoError(self):
  	if IGNORE_TEST:
  	  return
  	m = GAMES_PP(self.simple3)
  	m.analyze(error_details=False)
  	gr = GAMESReport(m)
  	som1 = m.getNode(self.simple3.getMolecule(CH3FH4))
  	som2 = m.getNode(self.simple3.getMolecule(FH4))
  	error = [[som1, som2]]
  	report, error_num = gr.reportTypeTwoError(error, explain_details=True)
  	self.assertEqual(error_num, [5])
  	LOC_START = 241
  	extended_report = NULL_STR
  	extended_report = extended_report + "{CH3FH4} < {CH2FH4=FFH2=FH2f=FH4} < {CH3FH4}\n\n"
  	extended_report = extended_report + "This indicates a mass conflict between reactions.\n"
  	extended_report = extended_report + "%s%s\n" % (PARAGRAPH_DIVIDER, PARAGRAPH_DIVIDER)
  	self.assertEqual(report[-LOC_START:], extended_report)

  def testConvertOperationSeriesToReactionOperations(self):
  	if IGNORE_TEST:
  	  return
  	m = GAMES_PP(self.simple4)
  	m.analyze(error_details=False)
  	gr = GAMESReport(m)
  	op = pd.Series([1.0, 0.5, 0.0], index = [STATPHOSPHORYLATION, PSTATDIMERISATION, PSTATDIMERISATIONNUC])
  	ro = gr.convertOperationSeriesToReactionOperations(op)
  	self.assertEqual(len(ro), 2)
  	self.assertEqual(ro[0].reaction, STATPHOSPHORYLATION)
  	self.assertEqual(ro[0].operation, 1.0)
  	self.assertEqual(ro[1].reaction, PSTATDIMERISATION)
  	self.assertEqual(ro[1].operation, 0.5)

  def testGetOperationMatrix(self):
    if IGNORE_TEST:
      return
    m1 = GAMES_PP(self.simple1)
    m1.analyze(error_details=False)
    gr1 = GAMESReport(m1)
    self.assertTrue(gr1.getOperationMatrix() is None)
    m4 = GAMES_PP(self.simple4)
    m4.analyze(error_details=False)
    gr4 = GAMESReport(m4)
    op_mat = gr4.getOperationMatrix()
    self.assertEqual(op_mat.loc[STATPHOSPHORYLATION, STATPHOSPHORYLATION], 1.0)
    self.assertEqual(op_mat.loc[PSTATDIMERISATION, PSTATDIMERISATION], 1.0)
    self.assertEqual(op_mat.loc[PSTATDIMERISATIONNUC, PSTATDIMERISATIONNUC], 1.0)
    self.assertEqual(op_mat.loc[PSTATDIMERISATIONNUC, STATPHOSPHORYLATION], 0.0)
    self.assertEqual(op_mat.loc[STATPHOSPHORYLATION, PSTATDIMERISATIONNUC], -0.5)

  def testGetResultingSeries(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple4)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    resulting_series = gr.getResultingSeries(STATPHOSPHORYLATION)
    print(resulting_series)
    self.assertEqual(resulting_series["{" + SPECIES_TEST + "}"], 1.0)
    self.assertEqual(resulting_series["{" + PSTAT_SOL + "}"], 0.0)
    self.assertEqual(resulting_series[m.getNode(m.simple.getMolecule(PSTATDIMER_NUC)).identifier], 0.0)
    self.assertEqual(resulting_series[m.getNode(m.simple.getMolecule(PSTAT_NUC)).identifier], 0.0)

  def testGetOperationStoichiometryMatrix(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple4)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    op = pd.Series([1.0, 0.5, 0.0], index = [STATPHOSPHORYLATION, PSTATDIMERISATION, PSTATDIMERISATIONNUC])
    ro = gr.convertOperationSeriesToReactionOperations(op)
    osm = gr.getOperationStoichiometryMatrix(ro)
    self.assertEqual(osm.loc[SPECIES_TEST, STATPHOSPHORYLATION], 1.0)
    self.assertEqual(osm.loc[SPECIES_TEST, PSTATDIMERISATION], 0.0)
    self.assertEqual(osm.loc[PSTAT_SOL, STATPHOSPHORYLATION], 1.0)
    self.assertEqual(osm.loc[PSTAT_SOL, PSTATDIMERISATION], -2.0)

  def testGeInferredReaction(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple4)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    op = pd.Series([1.0, 0.5, 0.0], index = [STATPHOSPHORYLATION, PSTATDIMERISATION, PSTATDIMERISATIONNUC])
    ro = gr.convertOperationSeriesToReactionOperations(op)
    inferred_reaction = gr.getInferredReaction(ro)
    self.assertEqual(len(inferred_reaction.reactants), 1)
    self.assertEqual(len(inferred_reaction.products), 2) 
    self.assertEqual(inferred_reaction.reactants[0].molecule.name, STAT_SOL)
    self.assertTrue(inferred_reaction.products[0].molecule.name in {PSTATDIMER_SOL, SPECIES_TEST})
    self.assertTrue(inferred_reaction.products[1].molecule.name in {PSTATDIMER_SOL, SPECIES_TEST})
    self.assertEqual(inferred_reaction.reactants[0].stoichiometry, 1.0)
    self.assertEqual({p.stoichiometry for p in inferred_reaction.products}, {0.5, 1.0})

  def testReportReactionsInSOM(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple4)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    som = m.getNode(m.simple.getMolecule(PSTATDIMER_NUC))
    report, error_num = gr.reportReactionsInSOM(som, 0)
    common_part = "1. PstatDimer__import: PstatDimer_sol -> PstatDimer_nuc\n"
    self.assertEqual(error_num, 1)
    self.assertTrue(report == common_part)

  def testReportEchelonError(self):
  	if IGNORE_TEST:
  	  return
  	m = GAMES_PP(self.simple4)
  	m.analyze(error_details=False)
  	gr = GAMESReport(m)
  	report, error_num = gr.reportEchelonError(m.echelon_errors, explain_details=True)
  	self.assertEqual(error_num, [3])
  	extended_report = NULL_STR
  	extended_report = extended_report + "will result in empty reactant with zero mass:\n\n:  -> {species_test}\n\n"
  	extended_report = extended_report + "\n----------------------------------------------------------------------\n"
  	extended_report = extended_report + "\n----------------------------------------------------------------------\n\n"
  	extended_report = extended_report + "\n\n**********************************************************************\n\n"
  	self.assertEqual(report[-288:], extended_report)

  def testReportTypeThreeError(self):
    if IGNORE_TEST:
      return
    m = GAMES_PP(self.simple4)
    m.analyze(error_details=False)
    gr = GAMESReport(m)
    report, error_num = gr.reportTypeThreeError(m.type_three_errors, explain_details=True)
    self.assertEqual(error_num, [3])
    pseudo_inequality_report = NULL_STR
    pseudo_inequality_report = pseudo_inequality_report + "6. statPhosphorylation: stat_sol -> Pstat_sol + species_test\n"
    pseudo_inequality_report = pseudo_inequality_report + "(pseudo 6.) statPhosphorylation: {Pstat_nuc=stat_nuc=stat_sol} -> "
    pseudo_inequality_report1 = pseudo_inequality_report + "{species_test} + {Pstat_sol}"
    pseudo_inequality_report2 = pseudo_inequality_report + "{Pstat_sol} + {species_test}"
    inference_report1 = "the masses of {Pstat_sol} and {Pstat_nuc=stat_nuc=stat_sol} are unequal."
    inference_report2 = "the masses of {Pstat_nuc=stat_nuc=stat_sol} and {Pstat_sol} are unequal."
    self.assertTrue(report[-460:-305]==pseudo_inequality_report1 or report[-460:-305]==pseudo_inequality_report2)
    self.assertTrue(report[-293:-221]==inference_report1 or report[-293:-221]==inference_report2)






"""
Tests for simple_sbml
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule
from SBMLLint.common import simple_sbml
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common import util

import numpy as np
import os
import libsbml
import unittest
import zipfile


IGNORE_TEST = False
NO_NAME = "dummy"
ANTIMONY_STG = '''
2S1 -> 3S2; 1
S1 = 0
S2 = 0
'''


#############################
# Tests
#############################
class TestSimpleSBML(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE)

  def testInitialize(self):
    if IGNORE_TEST:
      return
    simple = SimpleSBML()
    simple.initialize(cn.TEST_FILE)
    self.assertEqual(len(simple.reactions), cn.NUM_REACTIONS)
    self.assertEqual(len(simple.molecules), len(simple.moietys))

  def testGetMolecule(self):
    if IGNORE_TEST:
      return
    molecule1 = self.simple.molecules[0]
    name = molecule1.name
    molecule2 = self.simple.getMolecule(name)
    self.assertTrue(molecule1.isEqual(molecule2))
    self.assertIsNone(self.simple.getMolecule(NO_NAME))

  def testAdd(self):
    if IGNORE_TEST:
      return
    def test():
      self.assertEqual(len(self.simple.reactions), 2)
      for reaction in [reaction0, reaction1]:
        self.assertTrue(reaction in self.simple.reactions)
    #
    reaction0 = self.simple.reactions[0]
    reaction1 = self.simple.reactions[1]
    self.simple.reactions = [reaction0]
    self.simple.add(reaction1)
    test()
    self.simple.add(reaction1)
    test()

  def testRemove(self):
    num_reactions = len(self.simple.reactions)
    reaction0 = self.simple.reactions[0]
    reaction1 = self.simple.reactions[1]
    self.simple.remove(reaction0)
    self.assertTrue(reaction0 not in self.simple.reactions)
    self.assertTrue(reaction1 in self.simple.reactions)
    self.simple.add(reaction0)
    self.assertTrue(len(self.simple.reactions), num_reactions)

  def testGetReaction(self):
    if IGNORE_TEST:
      return
    reaction = self.simple.reactions[0]
    label = reaction.label
    reaction1 = self.simple.getReaction(label)
    self.assertTrue(reaction.isEqual(reaction1))


class TestFunctions(unittest.TestCase):

  def testReadURL(self):
    pass

  def _testIterator(self, itr):
     for item in itr:
       model = item.model
       self.assertTrue(isinstance(model.getSpecies(0),
           libsbml.Species))
     COUNT = 20
     itr = simple_sbml.modelIterator(final=COUNT)
     item_number = -1
     for item in itr:
       self.assertTrue(isinstance(item.filename, str))
       self.assertTrue(util.isSBMLModel(item.model))
       item_number = item.number
     self.assertEqual(item_number, COUNT - 1)

  def testModelIterator1(self):
    if IGNORE_TEST:
      return
    self._testIterator(simple_sbml.modelIterator(final=1))

  def testModelIterator2(self):
    if IGNORE_TEST:
      return
    self._testIterator(simple_sbml.modelIterator(
        final=1, zip_filename=None))

  def testGetZipfilePath(self):
    if IGNORE_TEST:
      return
    ffiles, zipper = simple_sbml.getZipfilePaths()
    for ffile in ffiles:
      try:
        fid = zipper.open(ffile)
        fid.close()
      except:
        assertTrue(False)
    

if __name__ == '__main__':
  unittest.main()



"""Set of Molecules(SOM)."""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML

from collections import deque

BRACKET_OPEN = "{"
BRACKET_CLOSE = "}"

class SOM(object):
  """
  The SOM (Set Of Molecules) class represents a collection of molecules
  each of which has equal weight. Consequently, the whole molecule space 
  can be partitioned into a collection of SOMs. The uni-uni reactions
  merge multiple SOM instances to create a larger one. 
  """
  def __init__(self, molecules, reactions=set()):
    """
    :param set-Molecule molecules:
    :param set-Reaction reactions:
    """
    self.molecules = molecules
    self.reactions = reactions
    self.identifier = self.makeId()

  def __repr__(self):
    return self.identifier        
      
  def makeId(self):
    """
    Creates an identifier for the SOM to uniquely
    identifies its elements.
    :return str:
    """
    def joinMoleculeNames(molecules):
      names = [m.name for m in molecules]
      names.sort()
      return '='.join(names)
    #
    identifier = "%s%s%s" % (
        BRACKET_OPEN, 
        joinMoleculeNames(list(self.molecules)),
        BRACKET_CLOSE
        )
    return identifier

  def merge(self, som):
    """
    Creates and returns a new SOM instance
    that has the union of molecules and reactions
    :param SOM som:
    :return SOM:
    """
    new_molecules = self.molecules.union(som.molecules)
    new_reactions = self.reactions.union(som.reactions)
    new_som = SOM(molecules=new_molecules, reactions=new_reactions)
    return new_som




#!/usr/bin/env python
"""
Runs the GAMES algorithm for a local XML file.
Usage: games <filepath>
"""

from SBMLLint.common import util
from SBMLLint.tools import sbmllint

import argparse

def main():
  parser = argparse.ArgumentParser(description='SBML XML file.')
  parser.add_argument('xml_file', type=open, help='SBML file')
  parser.add_argument('--config', type=open,
      help="SBMLLint configuration file")
  args = parser.parse_args()
  for fid in util.getNextFid(args.xml_file):
    try:
      sbmllint.lint(model_reference=fid,
          config_fid=args.config, mass_balance_check=sbmllint.GAMES)
    except ValueError:
      print ("  *** Bad SBML file.")


if __name__ == '__main__':
  main()



"""Access to SBMLLint configuration data.

Provides two functions.
  setConfiguration(<path>) - sets the value of the configuration dictionary.
  getConfiguration() - returns the current value of the configuration dictionary.
"""

import os
import yaml

import SBMLLint.common.constants as cn
from SBMLLint.common import msgs

_config_dict = None  # Configuration dictionary


def setConfiguration(path=cn.CFG_DEFAULT_PATH, fid=None):
  """
  :param str path: path to configuration file
  :param TextIOWrapper: overrides the path if present
  Updates _config_dict
  Notes:
    1. Changes string "True", "False" to booleans
    2. Inserts defaults for missing configuration keys
  """
  global _config_dict
  # Get the configuration file
  if fid is None:
    fid = open(path, "r")
  lines = fid.readlines()
  fid.close()
  lines = '\n'.join(lines)
  result = yaml.safe_load(lines)
  # Validate the section names
  for name in result.keys():
    if not name in cn.CFG_SECTIONS:
      msgs.error("Invalid section name: %s" % name)
  # Adjust values
  for k, v in result.items():
    if v == "True":
      result[k] = True
    if v == "False":
      result[k] = False
  for k, v in cn.CFG_DEFAULTS.items():
    if not k in result:
      result[k] = v 
  # Establish the configuration dictionary
  _config_dict = result

def getConfiguration():
  return _config_dict
#
setConfiguration(path=cn.CFG_DEFAULT_PATH)



"""
Tests for StoichiometryMatrix
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.games.som import SOM
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common.stoichiometry_matrix import StoichiometryMatrix

from scipy.optimize import linprog
import numpy as np
import os
import pandas as pd
import re
import unittest


IGNORE_TEST = False
REACTION1 = "reaction1"
REACTION4 = "reaction4"
MOLECULE_C = "C"
MOLECULE_M = "M"
REMAINIG_REACTIONS = 4
REMAINING_MOLECULES = 3


#############################
# Tests
#############################
class TestStoichiometryMatrix(unittest.TestCase):

  def setUp(self):
    self.simple_consistent = SimpleSBML()
    self.simple_consistent.initialize(cn.TEST_FILE8)
    self.consistent_matrix = StoichiometryMatrix(self.simple_consistent)
    #
    self.simple_inconsistent = SimpleSBML()
    self.simple_inconsistent.initialize(cn.TEST_FILE9)
    self.inconsistent_matrix = StoichiometryMatrix(self.simple_inconsistent)   

  def testConstructor(self):
    remaining_reactions = [r.label for r in self.consistent_matrix.reactions]
    print(remaining_reactions)
    remaining_molecules = self.consistent_matrix.molecules
    self.assertFalse(REACTION1 in remaining_reactions)
    self.assertTrue(REACTION4 in remaining_reactions)
    self.assertFalse(MOLECULE_C in remaining_molecules)
    self.assertTrue(MOLECULE_M in remaining_molecules)
    #
    self.assertEqual(len(self.inconsistent_matrix.reactions), REMAINIG_REACTIONS)
    self.assertEqual(len(self.inconsistent_matrix.molecules), REMAINING_MOLECULES)

  def testMakeStoichiometryMatrix(self):
    self.assertEqual(type(self.consistent_matrix), StoichiometryMatrix)
    self.assertEqual(type(self.inconsistent_matrix), StoichiometryMatrix)
    self.assertEqual(self.inconsistent_matrix.stoichiometry_matrix.shape, 
    	(REMAINING_MOLECULES, REMAINIG_REACTIONS))

  def testIsConsistent(self):
    self.assertTrue(self.consistent_matrix.consistent is None)
    self.assertTrue(self.inconsistent_matrix.consistent is None)
    self.assertTrue(self.consistent_matrix.isConsistent())
    self.assertFalse(self.inconsistent_matrix.isConsistent())

if __name__ == '__main__':
  unittest.main()
    


#!/usr/bin/env python
"""
Creates the moiety_structure section of the SBMLLint
configuration file. The user declares all moieties.
The moiety structure for a molecule is the number of
occurrences of each moiety string in the molecule.
Substrings are handled by looking first for the longest
moiety names.
"""

import argparse
import yaml
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common.moiety import MoietyStoichiometry
from SBMLLint.common import msgs


DEFAULT_CONFIG_FILE = "config.yml"
MOIETY_COUNT_SEPARATOR = ", "


def getMoieties(moiety_fid):
  """
  Creates the YAML configuration file.
  :param IOWrappter moiety_fid: 
  :return list-str: list of moiety names
  :raises ValueError: if one moiety is a substring of another
  """
  return yaml.safe_load(moiety_fid)

# TODO: Extend to do recursive checks of moiety names for
# their substructures
def findMoietyStoichiometries(molecule_name, moiety_names,
    is_check_error=True):
  """
  Finds the moieties and their stoichiometries for each molecule.
  Handles substrings by looking for largest strings first.
  :param str molecule:
  :param list-str moiety_names:
  :return list-MoietyStoichiometry:
  """
  result = []
  sorted_moiety_names = sorted(
      moiety_names, key=lambda n: len(str(n)), reverse=True)
  current_name = molecule_name
  for moiety_name in sorted_moiety_names:
    moiety_name = str(moiety_name)
    count = current_name.count(moiety_name)
    if count > 0:
      current_name = current_name.replace(moiety_name, "")
      result.append(MoietyStoichiometry(moiety_name, count,
          MOIETY_COUNT_SEPARATOR))
  current_name = current_name.replace("_", "")
  if is_check_error and len(current_name) > 0:
    msg = "Moieties specified do not completely express %s" % molecule_name
    msgs.error(msg)
  return result

# TODO: Look for moieties from largest string to smallest,
# eliminating those found.
def main(xml_fid, moiety_fid, config_fid):
  """
  Creates the YAML configuration file.
  :param IOWrapper xml_fid: xml file:
  :param IOWrapper moiety_fid: 
  :param IOWrapper config_fid: file to write
  """
  # Acquire and validate moieties
  moiety_names = getMoieties(moiety_fid)
  # Create the configuration file
  config_dct = {} 
  simple = SimpleSBML()
  simple.initialize(xml_fid)
  for molecule in simple.molecules:
    moiety_stoichiometries = findMoietyStoichiometries(
       molecule.name, moiety_names)
    config_dct[molecule.name] =  \
        [str(ms) for ms in moiety_stoichiometries]
  dct = {"moiety_structure": config_dct}
  yaml.dump(dct, config_fid)
  for fid in [xml_fid, moiety_fid, config_fid]:
    fid.close()

    
if __name__ == '__main__':
  def open_write(path):
    return open(path, "w")
  #
  parser = argparse.ArgumentParser(
      description='Creates configuration file for BioModels 147.')
  parser.add_argument('xml_file', 
      type=open, 
      help='SBML file from which species names are obtained.')
  parser.add_argument('moiety_file', 
      type=open, 
      help='YAML file the specifies moieties for the XML model.')
  parser.add_argument('--output', type=open_write,
      help="Generated configuration file. Defaults to config.yml.")
  args = parser.parse_args()
  if args.output is None:
    config_fid = open(DEFAULT_CONFIG_FILE, "w")
  else:
    config_fid = args.output
  main(args.xml_file, args.moiety_file, config_fid)



from SBMLLint.common import constants as cn
from SBMLLint.common import exceptions
from SBMLLint.common import util
import libsbml


import numpy as np
import os
import unittest


NUM_S1 = 2
NUM_S2 = 3
IGNORE_TEST = False
ANTIMONY_STG = '''
%dS1 -> %dS2; 1
S1 = 0
S2 = 0
''' % (NUM_S1, NUM_S2)
ZIP_PATH = os.path.join(cn.BIOMODELS_DIR, cn.BIOMODELS_ZIP_FILENAME)


#############################
# Tests
#############################
class TestFunctions(unittest.TestCase):

  def testGetXMLString(self):
    def test(xml):
      reader = libsbml.SBMLReader()
      document = reader.readSBMLFromString(xml)
      util.checkSBMLDocument(document)
      model = document.getModel()
      self.assertTrue('Reaction' in str(type(model.getReaction(0))))
    def getString(path):
      with open(path, 'r') as fd:
        lines = '\n'.join(fd.readlines())
      return lines
    #
    for path in [cn.TEST_FILE2, cn.TEST_FILE3]:
      try:
        test(util.getXML(path))
        test(util.getXML(getString(path)))
      except exceptions.MissingTelluriumError:
        pass

  def testGetXMLFromAntimony(self):
    try:
      xml = util.getXMLFromAntimony(ANTIMONY_STG)
    except exceptions.MissingTelluriumError:
      return
    self.assertTrue(isinstance(xml, str))
    reader = libsbml.SBMLReader()
    libsbml_document = reader.readSBMLFromString(xml)
    util.checkSBMLDocument(libsbml_document)
    model = libsbml_document.getModel()
    self.assertTrue('Reaction' in 
       str(type(model.getReaction(0))))

  def testIsInt(self):
    self.assertTrue(util.isInt(1))
    self.assertFalse(util.isInt(1.5))
    self.assertFalse(util.isInt('ab'))

  def testIsFloat(self):
    self.assertTrue(util.isFloat(1))
    self.assertTrue(util.isFloat(1.5))
    self.assertTrue(util.isFloat('1.5'))
    self.assertFalse(util.isFloat('ab'))

  def testIsSBMLModel(self):
    return
    self.assertFalse(util.isSBMLModel("dummy"))
    xml = util.getXML(cn.TEST_FILE2)
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromString(xml)
    util.checkSBMLDocument(document)
    model = document.getModel()
    self.assertTrue(util.isSBMLModel(model))

  def testUniqueify(self):
    class Tester():
    
      def __init__(self, name):
        self.name = name

      def __repr__(self):
        return self.name
 
      def isEqual(self, other):
        return self.name == other.name
    #
    STRING = 'abc'
    REPEATED_STRING = STRING + STRING   
    collection = [Tester(s) for s in REPEATED_STRING]
    result = util.uniqueify(collection)
    self.assertEqual(len(result), len(STRING))

  def testGetNextFid(self):
    fid = open(ZIP_PATH, "r")
    count = 0
    for zip_fid in util.getNextFid(fid, is_print=False):
      lines = zip_fid.read()
      count += 1
      self.assertGreater(len(lines), 0)
    self.assertGreater(count, 0)
    

if __name__ == '__main__':
  unittest.main()



"""
Tests for Reactions
"""
from SBMLLint.common import constants as cn
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction, REACTION_SEPARATOR

import numpy as np
import os
import unittest


IGNORE_TEST = False
NUM_S1 = 2
NUM_S2 = 3


#############################
# Tests
#############################
class TestReaction(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple.initialize(cn.TEST_FILE)
    self.reactions = self.simple.reactions
    self.reaction = self.reactions[2]

  def testGetId(self):
    identifier1 = self.reaction.getId()
    self.assertTrue(REACTION_SEPARATOR in identifier1)
    identifier2 = self.reaction.getId(is_include_kinetics=False)
    self.assertGreater(len(identifier1), len(identifier2))
    self.assertFalse(cn.KINETICS_SEPARATOR in identifier2)
    identifier3 = self.reaction.getId(is_include_kinetics=False,
        is_include_label=False)
    self.assertGreater(len(identifier2), len(identifier3))
    self.assertFalse(cn.LABEL_SEPARATOR in identifier3)

  def testConstructor(self):
    if IGNORE_TEST:
      return
    self.assertTrue(isinstance(self.reaction.reactants[0],
        MoleculeStoichiometry))
    self.assertTrue(isinstance(self.reaction.products[0],
        MoleculeStoichiometry))
    self.assertEqual(self.reaction.category, cn.REACTION_1_n)
    self.assertGreater(len(self.simple.molecules), 0)
    count = len(self.simple.reactions)
    reaction = self.simple.reactions[3]
    self.assertEqual(len(self.simple.reactions), count)

  # Test the stoichiometry
  def testConstructor2(self):
    if IGNORE_TEST:
      return
    simple = SimpleSBML()
    simple.initialize(cn.TEST_FILE2)
    reaction = simple.reactions[0]
    self.assertEqual(reaction.reactants[0].stoichiometry, NUM_S1)
    self.assertEqual(reaction.products[0].stoichiometry, NUM_S2)

  def testIsEqual(self):
    if IGNORE_TEST:
      return
    for reaction in self.simple.reactions[1:]:
      self.assertTrue(reaction.isEqual(reaction))
      self.assertFalse(reaction.isEqual(self.simple.reactions[0]))

  def testMakeIdentifier(self):
    if IGNORE_TEST:
      return
    for reaction in self.simple.reactions:
      parts = reaction.identifier.split('->')
      self.assertTrue(";" in parts[-1])  # Kinetics is last

  def testFindReactions(self):
    if IGNORE_TEST:
      return
    reactions = Reaction.find(self.reactions,
        category=cn.REACTION_1_n)
    trues = [r.category == cn.REACTION_1_n for r in reactions]
    self.assertTrue(all(trues))


if __name__ == '__main__':
  unittest.main()



"""Provides comparisons of moieties."""

from SBMLLint.common import constants as cn
from SBMLLint.common import config
from SBMLLint.common import util
from SBMLLint.common.reaction import Reaction
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common.moiety  \
    import Moiety, MoietyStoichiometry

import collections
import pandas as pd
import numpy as np

NULL_STR = ''
INDENT = "  "


MoietyComparatorResult = collections.namedtuple('MoietyComparatorResult',
    'num_reactions num_imbalances report')


class MoietyComparator(object):
  """Analysis of moieties of molecules."""

  def __init__(self, mol_stoichs1, mol_stoichs2, 
      names=["reactants", "products"]):
    """
    :param set-MoleculeStoichiometry mol_stoichs1:
    :param set-MoleculeStoichiometry mol_stoichs2:
    :param list-str names: names to refer to the two sets
    """
    def checkType(objs):
      trues = [isinstance(o, MoleculeStoichiometry)
          for o in objs]
      if all(trues):
        return
      raise ValueError("Argument must be collection of type %s"
          % str(MoleculeStoichiometry))
    checkType(mol_stoichs1)
    checkType(mol_stoichs2)
    config_dct = dict(config.getConfiguration())
    self._process_boundary_reactions =  \
        util.getKey(config_dct, cn.CFG_PROCESS_BOUNDARY_REACTIONS)
    if self._process_boundary_reactions is None:
      self._process_boundary_reactions = False
    self._ignored_molecules = util.setList(
        util.getKey(config_dct, cn.CFG_IGNORED_MOLECULES))
    self._ignored_moieties = util.setList(
        util.getKey(config_dct, cn.CFG_IGNORED_MOIETIES))
    self.molecule_stoichiometry_collections = [
        self._removeIgnoredMolecules(mol_stoichs1), 
        self._removeIgnoredMolecules(mol_stoichs2), 
        ]
    self.names = names

  def _removeIgnoredMolecules(self, mol_stoichs):
    """
    Removes ignored molecules from the molecule stoichmetries.
    :param list-MoleculeStoichiometry mol_stoichs:
    :return list-MoleculeStoichiometry:
    """
    names = [ms.molecule.name for ms in mol_stoichs]
    result = [ms for ms in mol_stoichs 
        if not ms.molecule.name in self._ignored_molecules]
    return result

  def _makeDFS(self):
    """
    Constructs a list of dataframes containing counts of moieties.
    :return list-pd.DataFrame: columns in DataFrame
        cn.VALUE
    """
    dfs = []
    for collection in self.molecule_stoichiometry_collections:
      if len(collection) > 0:
        df = MoleculeStoichiometry.countMoietysInCollection(collection)
        dfs.append(df)
    if len(dfs) == 0:
      dfs = [pd.DataFrame(), pd.DataFrame()]
    elif len(dfs) == 1:
      col = dfs[0].columns.tolist()[0]
      df = dfs[0].copy()
      df[col] = 0     
      dfs.append(df)
    return dfs

  def isSame(self):
    """
    Determines if the two molecules for the comparator
    have the same type and count of moieties.
    :return bool:
    """
    dfs = self._makeDFS()
    return dfs[0].equals(dfs[1])

  def difference(self):
    """
    Calculates the moieties present in one set and not in the other.
    :return pd.DataFrame: row index is moiety,
        column Value: counts of moieties present in 0 
        and not in second. Negative values are present in 1 and
        not in 0.
    Handles boundary reactions, when one set of moieties
    is all zeroes.
    """
    def addDFIndex(df, index):
      # Adds missing indices to df
      missing = set(index).difference(df.index)
      for item in missing:
        df.loc[item] = 0
    def isZeroColumn(df):
      col = df.columns[0]
      return df[col].sum() == 0
    #
    dfs = self._makeDFS()
    addDFIndex(dfs[0], dfs[1].index)
    addDFIndex(dfs[1], dfs[0].index)
    drops = set(self._ignored_moieties).intersection(dfs[0].index)
    df0 = dfs[0].drop(drops)
    df1 = dfs[1].drop(drops)
    df = df0 - df1
    # Handle boundaries
    if not self._process_boundary_reactions:
      if isZeroColumn(df0) or isZeroColumn(df1):
        df[df.columns[0]] = 0
    #
    return df

  def reportDifference(self):
    """
    Reports a difference in moieties between the two sets of molecules.
    :return str: report. null if no difference.
    """
    if self.isSame():
      return NULL_STR
    df = self.difference()
    #
    def appendNewline(stg):
      if len(stg) > 0:
        return "%s\n" % stg
      else:
        return stg
    #
    def buildStg(name, sign, indent=INDENT):
      """
      Constructs the report for the name with the given
      sign of values in df.
      :param str name:
      :param int sign: 1 or -1
      :param str indent: indentation on each line
      :return str:
      """
      stg = NULL_STR
      for idx, row in df.iterrows():
        if sign*row[cn.VALUE] > 0:
          stg = "%s%s%s: %2.2f\n" % (
              stg, indent, idx, sign*row[cn.VALUE])
      if not stg == NULL_STR:
        stg = "Excess moieties in %s\n%s" % (name, stg)
      return stg
    #
    stg1 = appendNewline(buildStg(self.names[0], 1))
    stg2 = appendNewline(buildStg(self.names[1], -1))
    return "%s%s" % (stg1, stg2)

  @classmethod
  def analyzeReactions(cls, model_reference):
    """
    Analyzes all reactions to detect moiety imbalances.
    :param libsbml.Model or SimpleSBML model:
    :return MoietyComparatorResult:
    If model_reference is SimpleSBML, it must have been initialized.
    """
    if isinstance(model_reference, SimpleSBML):
      simple = model_reference
    else:
      simple = SimpleSBML()
      simple.initialize(model_reference)
    num_imbalances = 0
    report = NULL_STR
    for reaction in simple.reactions:
      comparator = cls(reaction.reactants, reaction.products)
      stg = comparator.reportDifference()
      if len(stg) > 0:
        num_imbalances += 1
        report = "%s\n***%s\n%s" % (
            report, 
            reaction.getId(is_include_kinetics=False),
            stg
            )
    num_reactions = len(simple.reactions)
    report = "\n%d of %d reactions have imbalances.\n%s" % (
        num_imbalances, num_reactions, report)
    result = MoietyComparatorResult(
        num_reactions=num_reactions,
        num_imbalances=num_imbalances,
        report=report)
    return result



"""
Tests for Molecule
"""
from SBMLLint.common import constants as cn
from SBMLLint.common import config
from SBMLLint.common.moiety import Moiety, MoietyStoichiometry
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.simple_sbml import SimpleSBML

import itertools
import numpy as np
import os
import unittest


IGNORE_TEST = False
NUM1 = 2
NUM2 = 3
MOIETY_NAME1 = "first"
MOIETY_NAME2 = "second"
MOIETY_NAME3 = "third"
NAME = "name"
NO_NAME = 'not_a_name'
MOLECULE_NAME = "%s%s%s" % (MOIETY_NAME1, cn.MOIETY_DOUBLE_SEPARATOR, 
    MOIETY_NAME2)
MOLECULE_STOICHIOMETRY_STGS = {
    ((Moiety("A"), 1), (Moiety("P"), 1)): 
        [Molecule("A_P"), Molecule("A_1__P"), Molecule("A__P")],
    ((Moiety("AA"), 1), (Moiety("PP"), 1)): 
        [Molecule("AA_PP"), Molecule("AA_1__PP"), Molecule("AA__PP"),
        ],
    }
NAMES = [MOIETY_NAME1, MOIETY_NAME2, MOIETY_NAME3]
iterator = itertools.product([0,1], [0, 1], [0, 1])
MOLECULE_NAME_SET = []  # A set of names from moiety combinations
for item in iterator:
  name = ""
  for idx,ele in enumerate(item):
    if ele == 1:
      if len(name) == 0:
        name = NAMES[idx]
      else:
        name = "%s%s%s" % (name, cn.MOIETY_SEPARATOR, NAMES[idx])
  if len(name) > 0:
    MOLECULE_NAME_SET.append(name)
TEST_CFG_FILE = os.path.join(cn.TEST_DIR, "test_sbmllint_cfg.yml")


#############################
# Tests
#############################
class TestMolecule(unittest.TestCase):

  def _init(self):
    Molecule.molecules = []
    self.simple = SimpleSBML()
    self.simple = self.simple.initialize(cn.TEST_FILE)

  def setUp(self):
    if IGNORE_TEST:
      return
    self._init()

  def testConstructor(self):
    if IGNORE_TEST:
      return
    molecule = Molecule(NAME)
    self.assertEqual(molecule.name, NAME)

  def testAppend(self):
    if IGNORE_TEST:
      return
    moiety2 = Moiety(MOIETY_NAME2)
    molecule = Molecule(MOIETY_NAME1)
    new_molecule = molecule.append(moiety2)
    self.assertEqual(new_molecule.name, Molecule(MOLECULE_NAME).name)

  def testGetMoietys(self):
    if IGNORE_TEST:
      return
    m_s1 = MoietyStoichiometry(Moiety(MOIETY_NAME1), NUM1)
    m_s2 = MoietyStoichiometry(MOIETY_NAME2, NUM2)
    m_s3 = MoietyStoichiometry(MOIETY_NAME1, NUM2)
    molecule = Molecule(str(m_s1))
    molecule = molecule.append(Moiety(str(m_s2)))
    moietys = molecule.getMoietys()
    expected = [Moiety(MOIETY_NAME1), Moiety(MOIETY_NAME2)]
    for moiety in moietys:
      self.assertTrue(any([moiety.isEqual(e) for e in expected]))

  def testHasMoiety(self):
    if IGNORE_TEST:
      return
    molecule = Molecule(MOLECULE_NAME)
    self.assertTrue(molecule.hasMoiety(Moiety(MOIETY_NAME1)))
    self.assertTrue(molecule.hasMoiety(Moiety(MOIETY_NAME2)))

  def testMoietyStoichiometrys(self):
    if IGNORE_TEST:
      return
    config.setConfiguration(TEST_CFG_FILE)
    #
    molecule = Molecule("Glu6P")
    self.assertEqual(molecule.moiety_stoichiometrys[0].moiety.name,
        "Glu")
    #
    molecule = Molecule("ATP")
    self.assertEqual(len(molecule.moiety_stoichiometrys), 2)


class TestMoleculeStoichiometry(unittest.TestCase):

  def setUp(self):
    self.simple = SimpleSBML()
    self.simple = self.simple.initialize(cn.TEST_FILE)
    Molecule.molecules = []

  def testConstructor(self):
    if IGNORE_TEST:
      return
    molecule = Molecule(MOLECULE_NAME)
    m_s = MoleculeStoichiometry(molecule, NUM1)
    self.assertTrue(m_s.molecule.isEqual(molecule))
    self.assertEqual(m_s.stoichiometry, NUM1)
  
  def testExtractMoietyStoichiometrys(self):
    if IGNORE_TEST:
      return
    for expected, strings in MOLECULE_STOICHIOMETRY_STGS.items():
      moietys = list(set([e[0] for e in expected]))
      moietys.sort()
      m_ss = [MoietyStoichiometry(m, 1) for m in moietys]
      for stg in strings:
        result = Molecule.getMoietyStoichiometrys(stg)
        trues = [x.isEqual(y) for x, y in zip(result, m_ss)]
        self.assertTrue(all(trues))

  def testExtractMoietyStoichiometrys(self):
    if IGNORE_TEST:
      return
    moiety_stoich = MoietyStoichiometry(Molecule(MOIETY_NAME1),
        NUM1)
    mole_stoich = MoleculeStoichiometry(Molecule(str(moiety_stoich)), 
        NUM2)

  def testCountMoietys(self):
    if IGNORE_TEST:
      return
    moiety_stoich = MoietyStoichiometry(Molecule(MOIETY_NAME1),
        NUM1)
    mole_stoich = MoleculeStoichiometry(Molecule(str(moiety_stoich)), 
        NUM2)
    df = mole_stoich.countMoietys()
    self.assertEqual(df.columns.tolist(), [cn.VALUE])
    expected = NUM1 * NUM2
    trues = [expected ==  n for n in df[cn.VALUE]]

  def testCountMoietys2(self):
    if IGNORE_TEST:
      return
    m_ss = [
        MoleculeStoichiometry(Molecule("A_P_P_P"), 1),
        MoleculeStoichiometry(Molecule("A__P_3"), 1),
        ]
    dfs = []
    for m_s in m_ss:
      dfs.append(m_s.countMoietys())
    self.assertTrue(dfs[0].equals(dfs[1]))

  def testCountMoietysInCollection(self):
    if IGNORE_TEST:
      return
    m_ss = [MoleculeStoichiometry(Molecule(n), NUM1)
        for n in MOLECULE_NAME_SET[:3]]
    df = MoleculeStoichiometry.countMoietysInCollection(m_ss)
    indexes = df.index.tolist()
    self.assertEqual(len(indexes), len(set(indexes)))

  def testGetMolecules(self):
    if IGNORE_TEST:
      return
    names = ["a", "b", "c"]
    full_names = list(names)
    full_names.extend(names)
    m_ss = [MoleculeStoichiometry(Molecule(n), NUM1)
        for n in full_names]
    molecules = MoleculeStoichiometry.getMolecules(m_ss)
    self.assertEqual(len(molecules), len(names))
    

if __name__ == '__main__':
  unittest.main()



"""Commonly used utilities."""

from SBMLLint.common import constants as cn
from SBMLLint.common.tellurium_sandbox import TelluriumSandbox

import os
import zipfile

TYPE_ANTIMONY = "type_antimony"
TYPE_XML = "type_xml"
TYPE_FILENAME = "type_filename"
XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>'

def getXML(model_reference):
  """
  :param str model_reference: 
      the input may be a file reference or a model string
      or TextIOWrapper
          and the file may be an xml file or an antimony file.
      if it is a model string, it may be an xml string or antimony.
  :raises IOError: Error encountered reading the SBML document
  :return str SBML xml"
  """
  # Check for a file path
  model_str = ""
  if isinstance(model_reference, str):
    if os.path.isfile(model_reference):
      with open(model_reference, 'r') as fd:
        lines = fd.readlines()
      model_str = ''.join(lines)
  if len(model_str) == 0:
    if "readlines" in dir(model_reference):
      lines = model_reference.readlines()
      if isinstance(lines[0], bytes):
        lines = [l.decode("utf-8") for l in lines]
      model_str = ''.join(lines)
      model_reference.close()
    else:
      # Must be a string representation of a model
      model_str = model_reference
  # Process model_str into a model  
  if not "<sbml" in model_str:
    # Antimony
    model_str = getXMLFromAntimony(model_str)
  return model_str

def getXMLFromAntimony(antimony_stg):
  """
  Constructs an SBML model from the antimony string.
  :param str antimony_stg:
  :return str: SBML model in xml format
  """
  sandbox = TelluriumSandbox()
  sandbox.run("getSBMLFromAntimony", antimony_stg)
  if sandbox.return_code != 0:
    raise ValueError("Bad antimony string: %s" % antimony_stg)
  return sandbox.output

def isInt(obj):
  try:
    return str(int(obj)) == str(obj)
  except:
    return False

def isFloat(obj):
  try:
    value = float(obj)
  except:
    return False
  return True

def isSBMLModel(obj):
  """
  Tests if object is a libsbml model
  """
  cls_stg = str(type(obj))
  if ('Model' in cls_stg) and ('lib' in cls_stg):
    return True
  else:
    return False

def uniqueify(collection):
  """
  Prunes the collection so that only unique objects are present.
  Elements of the collection must have the method "isEqual" that
  takes as an argument another member of the collection.
  :param list-obj collection
  :return list-obj:
  """
  result = []
  for ele in collection:
    if all([not ele.isEqual(r) for r in result]):
      result.append(ele)
  return result
     
def checkSBMLDocument(document, model_reference=""): 
  if (document.getNumErrors() > 0):
    raise ValueError("Errors in SBML document\n%s" 
        % model_reference)

def setList(a_list):
  if a_list is None:
    return []
  else:
    return a_list

def getKey(dct, key):
  """
  Returns a value if the key is present or None.
  """
  if key in dct.keys():
    return dct[key]
  else:
    return None

def getNextFid(fid, is_print=True):
  """
  Iterator for files in a zip archive.
  If fid is not a zipfile, then just returns that fid.
  :param IOTextWrapper fid:
  :param bool is_print: prints the file name
  :return fid:
  Usage: for zip_fid in getNextFid(fid):
  """
  path = fid.name
  splits = os.path.splitext(path)
  if splits[1] != ".zip":
    yield fid
  else:
    # Zip file
    fid.close()  # Need to open as a zipfile
    with zipfile.ZipFile(path, "r") as zipper:
      for ffile in zipper.filelist:
        zip_fid = zipper.open(ffile, "r")
        if is_print:
          print("\n** %s" % zip_fid.name)
        yield zip_fid



"""Mass Equality Set Graph (MESGraph)."""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.games.som import SOM
from SBMLLint.common.simple_sbml import SimpleSBML

import collections
import itertools
import networkx as nx

NULL_STR = ""


MESGraphReport = collections.namedtuple("MESGraphReport", 
    "type_one type_two type_three type_four type_five")
SOMMoleculeStoichiometry = collections.namedtuple("SOMStoichiometry",
    "som molecule stoichiometry")

class MESGraph(nx.DiGraph):
  """
  The MESGraph class represents a collection of SOMs as nodes
  and their inequality relationships as edges (arcs).
  Mass inequality between SOMs from reactions can help us
  detect their relationship.
  Type I Error occurs when we find inequality between two molecules
  in the same SOM, because each element in a SOM has the same weight.
  Type II Error implies there is cyclism between molecules, such as
  A < B < C < ... < A, which is physically impossible.
  """

  def __init__(self, simple=None):
    """
    :param SimpleSBML simple:
    """
    super(MESGraph, self).__init__()
    self.simple = simple
    self.soms = self.initializeSOMs(simple)
    self.add_nodes_from(self.soms)
    self.identifier = self.makeId()
    self.multimulti_reactions = []
    self.type_one_error = False
    self.type_two_error = False
    self.type_one_errors = []
    self.type_two_errors = []
    self.type_three_errors = []
    self.type_four_errors = []
    self.type_five_errors = []

  def __repr__(self):
    return self.identifier

  def initializeSOMs(self, simple):
    """
    Create a list of one-molecule SOMs
    :param SimpleSBML simple:
    :return list-SOM:
    """
    soms = []
    if type(simple) == SimpleSBML:
      for molecule in simple.molecules:
        if molecule.name == cn.EMPTYSET:
          continue
        else:
          soms.append(SOM({molecule}))
    return soms

  def makeId(self):
    """
    Construct an identifier for the graph.
    :return str:
    """
    identifier = ""
    if self.edges:
      for edge in self.edges:
        identifier = identifier + str(edge[0]) + cn.ARC_ARROW + str(edge[1]) + "\n"
    for key, node in enumerate(nx.isolates(self)):
      identifier = identifier + str(node)
      if key < (len(list(nx.isolates(self)))-1):
          identifier = identifier + cn.KINETICS_SEPARATOR
    # Return the identifier
    return identifier

  def getNode(self, molecule):
    """
    Find a node(SOM) containing the given molecule.
    If no such SOM exists, return False
    :param Molecule molecule:
    :return SOM/False:
    """
    for som in list(self.nodes):
      for mole in som.molecules:
        if mole.name == molecule.name:
          return som
    return False

  def mergeNodes(self, som1, som2, reaction):
    """
    Merge two nodes (SOMs).
    Update arcs if applicable. 
    :param SOM som1:
    :param SOM som2:
    :param Reaction reaction:
    :return SOM new_som:
    """
    new_som = som1.merge(som2)
    new_som.reactions.add(reaction)
    for som in [som1, som2]:
      for edge in list(self.in_edges(som)):
        remaining_som = edge[0]
        reaction_label = self.get_edge_data(edge[0], edge[1])[cn.REACTION]
        self.add_edge(remaining_som, new_som, reaction=reaction_label)  
      for edge in list(self.out_edges(som)):
        remaining_som = edge[1]
        reaction_label = self.get_edge_data(edge[0], edge[1])[cn.REACTION]
        self.add_edge(new_som, remaining_som, reaction=reaction_label) 
    self.remove_nodes_from([som1, som2])
    if not self.has_node(new_som):
      self.add_node(new_som)
    return new_som

  def processUniUniReaction(self, reaction):
    """
    Process a 1-1 reaction to merge nodes.
    If no need to merge, return None.
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_1_1:
      pass
    else:
      reactant_som = self.getNode(reaction.reactants[0].molecule)
      product_som = self.getNode(reaction.products[0].molecule)
      if reactant_som == product_som:
        return None
      else:
        # new_som = reactant_som.merge(product_som)
        # new_som.reactions.add(reaction)
        # # TODO: if there are edges, need to also check them: self.mergeNodes(som1, som2, reaction)
        # self.remove_node(reactant_som)
        # self.remove_node(product_som)
        # self.add_node(new_som)
        new_som = self.mergeNodes(reactant_som, product_som, reaction)
        self.identifier = self.makeId()
        return new_som

  def processUniMultiReaction(self, reaction):
    """
    Process a 1-n reaction to add arcs.
    Since the mass of reactant is greater than
    that of each product, it adds arcs by
    addArc(source=products, destination=reactant).
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_1_n:
      pass
    else:
      destination = [reaction.reactants[0].molecule]
      source = [product.molecule for product in reaction.products]
      arcs = itertools.product(source, destination)
      for arc in arcs:
        if not self.checkTypeOneError(arc, reaction):
          som_source = self.getNode(arc[0])
          som_destination = self.getNode(arc[1])
          self.addArc(som_source, som_destination, reaction)
      self.identifier = self.makeId()

  def processMultiUniReaction(self, reaction):
    """
    Process a n-1 reaction to add arcs.
    Since the mass of product is greater than
    that of each reactant, it adds arcs by
    addArc(source=reactants, destination=product).
    :param Reaction reaction:
    """
    if reaction.category != cn.REACTION_n_1:
      pass
    else:
      destination = [reaction.products[0].molecule]
      source = [reactant.molecule for reactant in reaction.reactants]
      arcs = itertools.product(source, destination)
      for arc in arcs:
        if not self.checkTypeOneError(arc, reaction):
          som_source = self.getNode(arc[0])
          som_destination = self.getNode(arc[1])
          self.addArc(som_source, som_destination, reaction)
      self.identifier = self.makeId()

  def addMultiMultiReaction(self, reaction=None):
    """
    Add a multi-multi reaction to self.multimulti_reactions
    :param reaction Reaction:
    :return bool:
    """
    if reaction not in self.multimulti_reactions:
      self.multimulti_reactions.append(reaction)

  def addTypeThreeError(self, som1, som2, reaction):
    """
    Add Type III Error components to self.type_three_errors
    All components of resulting PathComponents are str
    :param SOM som1:
    :param SOM som2:
    :param Reaction reaction:
    :return bool flag:
    """
    flag = False
    for component in self.type_three_errors:
      if (component.node1==som1) and (component.node2==som2):
        new_component = cn.PathComponents(node1=som1, 
                                          node2=som2,
                                          reactions=component.reactions+[reaction.label])
        self.type_three_errors.remove(component)
        self.type_three_errors.append(new_component)
        flag = True
        break
    if not flag:
      self.type_three_errors.append(cn.PathComponents(node1=som1, 
                                                      node2=som2,
                                                      reactions=[reaction.label]))
      flag = True
    return flag  

  def checkTypeThreeError(self, som1, som2, reaction):
    """
    Check type III error, which is when
    we cannot merge two nodes because there is an arc.
    Add the error to type three error. 
    :param SOM som1:
    :param SOM som2:
    :param reaction Reaction:
    :return bool:
    """
    if self.has_edge(som1, som2):
      self.addTypeThreeError(som1, som2, reaction)
      return True
    elif self.has_edge(som2, som1):
      self.addTypeThreeError(som2, som1, reaction)
      return True
    else:
      return False

  def reduceReaction(self, reaction):
    """
    Reduce the given reaction 
    :param Reaction reaction:
    :return False/Reaction reaction:
    """
    if reaction.category != cn.REACTION_n_n:
        return False
    # Reduces the reaction by examining for each SOM
    for som in list(self.nodes):
      reactants_in = collections.deque([mole_stoich for mole_stoich in  
                           reaction.reactants if 
                           self.getNode(mole_stoich.molecule)==som])
      reactants_out = [mole_stoich for mole_stoich in  
                      reaction.reactants if 
                      self.getNode(mole_stoich.molecule)!=som]
      products_in = collections.deque([mole_stoich for mole_stoich in  
                          reaction.products if 
                          self.getNode(mole_stoich.molecule)==som])
      products_out = [mole_stoich for mole_stoich in  
                     reaction.products if 
                      self.getNode(mole_stoich.molecule)!=som]
      #
      while reactants_in and products_in:
        reactant = reactants_in[0]
        product = products_in[0]
        if reactant.stoichiometry > product.stoichiometry:
          reactants_in[0] = MoleculeStoichiometry(reactant.molecule,
                                                  reactant.stoichiometry - product.stoichiometry)
          products_in.popleft()
        elif reactant.stoichiometry < product.stoichiometry:
          products_in[0] = MoleculeStoichiometry(product.molecule,
                                                 product.stoichiometry - reactant.stoichiometry)
          reactants_in.popleft()
        else:
          reactants_in.popleft()
          products_in.popleft()
      reactants = list(reactants_in) + reactants_out
      products = list(products_in) + products_out
      #  
      if (len(reaction.reactants) > len(reactants)) | \
        (len(reaction.products) > len(products)):
        reaction.reactants = reactants
        reaction.products = products
    reaction.identifier = reaction.makeIdentifier()
    reaction.category = reaction.getCategory() 
    # reduced_reaction = cn.ReactionComponents(label = reaction.label, reactants=reactants, products=products)
    return reaction

  # def addTypeFourError(self, reduced_reaction):
  #   """
  #   Add type IV error, which is when reduced reaction
  #   only has one elements on one side.
  #   :param Reaction reduced_reaction:
  #   :return bool:
  #   """
  #   if som1 == som2:
  #     self.addTypeFourError(som1, som2, reaction)
  #     return True
  #   return False

  def processMultiMultiReaction(self, reaction):
    """
    Process a multi-multi reaction.
    Return False means graph was updated or error was added
    Return True means graph wasn't updated 
    and error wasn't added (already exist or couldn't find)
    :param Reaction reaction:
    :return bool:
    """
    # Reduce the reaction using existing SOMs
    reduced_reaction = self.reduceReaction(reaction)
    # If the reaction was not a multi-multi reaction, return False
    if not reduced_reaction:
      return False
    # if reduced reaction is EmptySet -> EmptySet, don't do anything
    if len(reduced_reaction.reactants)==0 and len(reduced_reaction.products)==0:
      return False
    # elif reduced reaction has exactly one side EmptySet, add type four error
    elif len(reduced_reaction.reactants)==0 or len(reduced_reaction.products)==0:
      if reduced_reaction in self.type_four_errors:
        return False
      else:
        self.type_four_errors.append(reduced_reaction)
        return True
    reactant_soms = list({self.getNode(ms.molecule) for ms in reduced_reaction.reactants})
    product_soms = list({self.getNode(ms.molecule) for ms in reduced_reaction.products})
    reactant_stoichiometry = [ms.stoichiometry for ms in reduced_reaction.reactants]
    product_stoichiometry = [ms.stoichiometry for ms in reduced_reaction.products]
    sum_reactant_stoichiometry = sum(reactant_stoichiometry)
    sum_product_stoichiometry = sum(product_stoichiometry)
    #
    # if both sides have more than one SOMS, we cannot process (return False)
    # if (len(reactant_soms)>1) and (len(product_soms)>1):
    #   return False
    # if both sides have exactly one SOM
    if (len(reactant_soms)==1) and (len(product_soms)==1):
      reactant_som = reactant_soms[0]
      product_som = product_soms[0]
      if sum_reactant_stoichiometry == sum_product_stoichiometry:
        if reactant_som != product_som:
          if not self.checkTypeThreeError(reactant_som, product_som, reaction):
            self.mergeNodes(reactant_som, product_som, reaction)
      # Add reactant_som -> product_som
      elif sum_reactant_stoichiometry > sum_product_stoichiometry:
        self.addArc(reactant_som, product_som, reaction)
      # Add product_som -> reactant_som
      else:
        self.addArc(product_som, reactant_som, reaction)
      self.identifier = self.makeId()
      return True
    # if one side has exactly one SOM, and the other side multiple SOMs
    else: 
      # SOM uni-multi reaction
      if (len(reactant_soms)==1) and \
          (sum_reactant_stoichiometry==1) and \
          (len(product_soms)>1):
        som_arcs = itertools.product(product_soms, reactant_soms)
      # SOM multi-uni reaction
      elif (len(reactant_soms)>1) and \
          (len(product_soms)==1) and \
          (sum_product_stoichiometry==1):
        som_arcs = itertools.product(reactant_soms, product_soms)
      # The rest are all multi-multi 
      # that is, multiple SOMs on both sides, or one side one SOM with stoichiometry>1
      else:
        return self.processByInequality(reduced_reaction)
      for arc in som_arcs:
        self.addArc(arc[0], arc[1], reaction)
      self.identifier = self.makeId()
      return True
    # return none if none of above applied (should not happen)
    return None

  def processByInequality(self, reduced_reaction):
    """
    Cross-examine inequality and add arcs if possible.
    One example is, A + B -> C + D where A < C is given
    by existing arc. 
    We can conclude B > D and add an arc.  
    Return True if processed, False otherwise
    :param Reaction reduced_reaction:
    :return bool:
    """  
    # create lists of soms with stoichiometry
    reactants = collections.deque([
        SOMMoleculeStoichiometry(som = self.getNode(moles.molecule),
            molecule = moles.molecule, 
            stoichiometry = moles.stoichiometry) for \
            moles in reduced_reaction.reactants])
    products = collections.deque([
        SOMMoleculeStoichiometry(som = self.getNode(moles.molecule), 
            molecule = moles.molecule,
            stoichiometry = moles.stoichiometry) for \
            moles in reduced_reaction.products])
    reactant_soms = list({reactant.som for reactant in reactants})
    product_soms = list({product.som for product in products})
    #
    reactant_lessthan_product = []
    product_lessthan_reactant = []
    no_relationship = []
    for pair in itertools.product(reactant_soms, product_soms):
      if self.has_edge(pair[0], pair[1]):
        reactant_lessthan_product.append(pair)
      elif self.has_edge(pair[1], pair[0]):
        product_lessthan_reactant.append(pair)
      else:
        no_relationship.append(pair)
    # print("reduced reaction...", reduced_reaction.makeIdentifier(is_include_kinetics=False))
    # print("reactant_lessthan_product: ", reactant_lessthan_product)
    # print("product_lessthan_reactant: ", product_lessthan_reactant)
    # print("no_realtionship :", no_relationship)
    # print("----------------------------------------------------------")
    # now, want to infer the relationship of no_relationship
    # or prove if existing relationships conflict
    if not no_relationship:
      return False
    # if both directions exist, let's say we cannot do anything; return False
    if reactant_lessthan_product and product_lessthan_reactant:
      return False
    def processPairs(pairs, small, big, idx_small, idx_big):
    # under product_lessthan_reactant, idx_small = 1, idx_big = 0
    # under the same, small = products, big = reactants
    # soms_buffer is same side as small_som
    # remaining_soms is same side as big_som
      big_som_stoichiometry = 0
      small_som_stoichiometry = 0
      soms_buffer = [pair[idx_small] for pair in no_relationship]
      remaining_soms = [pair[idx_big] for pair in no_relationship]
      for pair in pairs:
        print("We are dealing with, ", pair)
        big_som_stoichiometry += sum([
            sms.stoichiometry for sms in big if sms.som==pair[idx_big]])
        small_som_stoichiometry += sum([
            sms.stoichiometry for sms in small if sms.som==pair[idx_small]])
        if pair[idx_small] in soms_buffer:
          soms_buffer.remove(pair[idx_small])
        if pair[idx_big] in remaining_soms:
          remaining_soms.remove(pair[idx_big])
      print("big_som_stoi, ", big_som_stoichiometry)
      print("small_som_stoi, ", small_som_stoichiometry)
      # if product_som_stoichiometry is bigger, it's okay
      # if not, check if there is at least one buffer on product; 
      # if yes, try adding an arc if the buffer is at least one
      if big_som_stoichiometry < small_som_stoichiometry:
        return False
      elif soms_buffer:
        if len(soms_buffer)==1:
          # add arc
          for arc_source in remaining_soms:
            # the SOMs cannot be the same because they were already reduced
            self.addArc(arc_source, soms_buffer[0])
            return True
        # cannot decide now because there are more than two buffers
        else:
          return False
      # no buffer; add error
      else:
        if reduced_reaction in self.type_four_errors:
          return False
        else:
          self.type_four_errors.append(reduced_reaction)
          print("type four error added!", reduced_reaction)
          return True
    if product_lessthan_reactant:
      return processPairs(
          pairs=product_lessthan_reactant, 
          big=reactants, 
          small=products, 
          idx_big=0, idx_small=1)
    elif reactant_lessthan_product:
      return processPairs(
          pairs=reactant_lessthan_product, 
          big=products, 
          small=reactants, 
          idx_big=1, idx_small=0)
    return False

  def addArc(self, arc_source, arc_destination, reaction):
    """
    Add a single arc (edge) using two SOMs and reaction.
    :param SOM arc_source:
    :param SOM arc_destination:
    :param Reaction reaction:
    """
    # if there is already a preious reaction,
    if self.has_edge(arc_source, arc_destination):
      reaction_label = self.get_edge_data(arc_source, arc_destination)[cn.REACTION]
      # if reaction.label is not already included in the attribute,
      if reaction.label not in set(reaction_label):
        reaction_label = reaction_label + [reaction.label]
    else:
      reaction_label = [reaction.label]
    # overwrite the edge with new reactions set
    self.add_edge(arc_source, arc_destination, reaction=reaction_label)

  def getSOMPath(self, som, mole1, mole2):
    """
    Create an undirected graph between
    two molecules within a SOM
    and find the shortest path
    :param SOM som:
    :param Molecule mole1:
    :param Molecule mole2:
    :return PathComponents som_path:
    """   
    molecule1 = mole1.name
    molecule2 = mole2.name
    # construct undirected graph
    subg = nx.Graph()
    # here, every reaction is 1-1 reaction
    for reaction in list(som.reactions):
      node1 = reaction.reactants[0].molecule.name
      node2 = reaction.products[0].molecule.name
      if subg.has_edge(node1, node2):
        reaction_label = subg.get_edge_data(node1, node2)[cn.REACTION]
        # if reaction.label is not already included in the attribute,
        if reaction.label not in set(reaction_label):
          reaction_label = reaction_label + [reaction.label]
      else:
        reaction_label = [reaction.label]    
      subg.add_edge(node1, node2, reaction=reaction_label)
    path = [short_p for short_p in nx.shortest_path(subg, 
                                                    source=molecule1, 
                                                    target=molecule2)]
    som_path = []
    for idx in range(len(path)-1):
      edge_reactions = subg.get_edge_data(path[idx], path[idx+1])[cn.REACTION]
      som_path.append(cn.PathComponents(node1=path[idx], 
                                     node2=path[idx+1],
                                     reactions=edge_reactions))
    return som_path

  def printSOMPath(self, molecule_name1, molecule_name2):
    """
    Print out shortest SOM path between two molecules.
    Arguments are str and both molecules sholud be in the 
    same SOM.
    :param str molecule_name1:
    :param str molecule_name2:
    :return bool/str:
    """
    path_report = NULL_STR
    som1 = self.getNode(self.simple.getMolecule(molecule_name1))
    som2 = self.getNode(self.simple.getMolecule(molecule_name2))
    if som1 != som2:
      return False
    else:
      # add case when molecule_name1 == molecule_name2
      if molecule_name1 == molecule_name2:
        # print("Clearly,", molecule_name1, cn.EQUAL, molecule_name2)
        path_report = path_report + "Clearly, %s %s %s\n" % (
            molecule_name1, cn.EQUAL, molecule_name2)
      else:
        som_path = self.getSOMPath(som1, 
                                   self.simple.getMolecule(molecule_name1), 
                                   self.simple.getMolecule(molecule_name2))
        for pat in som_path:
          # print("\n%s %s %s by reaction(s):" % (pat.node1, cn.EQUAL, pat.node2))
          path_report = path_report + "\n%s %s %s by reaction(s):\n" % (pat.node1, cn.EQUAL, pat.node2)
          for r in pat.reactions:
            som_reaction = self.simple.getReaction(r)
            # print(som_reaction.makeIdentifier(is_include_kinetics=False))
            path_report = path_report + "%s\n" % (som_reaction.makeIdentifier(is_include_kinetics=False))
      return path_report

  def addTypeOneError(self, mole1, mole2, reaction):
    """
    Add Type I Error components to self.type_one_errors
    All components of resulting PathComponents are str
    :param Molecule mole1:
    :param Molecule mole2:
    :param Reaction reaction:
    :return bool flag:
    """
    flag = False
    for component in self.type_one_errors:
      if (component.node1==mole1.name) and (component.node2==mole2.name):
        new_component = cn.PathComponents(node1=mole1.name, 
                                          node2=mole2.name,
                                          reactions=component.reactions+[reaction.label])
        self.type_one_errors.remove(component)
        self.type_one_errors.append(new_component)
        flag = True
        break
    if not flag:
      self.type_one_errors.append(cn.PathComponents(node1=mole1.name, 
                                                    node2=mole2.name,
                                                    reactions=[reaction.label]))
      flag = True
    return flag

  def checkTypeOneError(self, arc, inequality_reaction=None):
    """
    Check Type I Error of an arc.
    If both source and destination are found
    in the same SOM, send error message and return True.
    If not, return False.
    :param tuple-Molecule arc:
    :param Reaction inequality_reaction:
    :return bool:
    """
    som1 = self.getNode(arc[0])
    som2 = self.getNode(arc[1])
    if som1 == som2:
      self.addTypeOneError(arc[0], arc[1], inequality_reaction)
      return True
    else:
      return False

  def addTypeTwoError(self, cycle):
    """
    Add Type II Error components to self.type_two_errors
    which is a list of lists
    All components of resulting PathComponents are str
    :param list-SOM cycle:
    """
    # exceptionally, here PathComponents are
    # node1=[], node2=[], reactions=[] and their index
    # of each component will match. All elements within nodes
    # are in the same SOM
    error_cycle = []
    for node_idx in range(len(cycle)-1):
      som1 = cycle[node_idx]
      som2 = cycle[node_idx+1]
      som1_moles = {mole.name for mole in list(som1.molecules)}
      som2_moles = {mole.name for mole in list(som2.molecules)}
      reactions = self.get_edge_data(som1, som2)[cn.REACTION]
      # all reactions (in an edge), should create a single PathComponent
      nodes1 = []
      nodes2 = []
      reaction_labels = []
      for r in reactions:
        reaction = self.simple.getReaction(r)
        if reaction.category == cn.REACTION_n_1:
          sources = {r.molecule.name for r in reaction.reactants}
          destinations = {p.molecule.name for p in reaction.products}
        elif reaction.category == cn.REACTION_1_n:
          sources = {p.molecule.name for p in reaction.products}
          destinations = {r.molecule.name for r in reaction.reactants}
        # for any reaction that addes arcs, len(nodes2)==1
        node2 = list(destinations.intersection(som2_moles))[0]
        for node1 in list(sources.intersection(som1_moles)):
          nodes1.append(node1)
          nodes2.append(node2)
          reaction_labels.append(reaction.label)
    error_cycle.append(cn.PathComponents(node1=nodes1, 
                                         node2=nodes2,
                                         reactions=reaction_labels))
    som1 = cycle[-1]
    som2 = cycle[0]
    som1_moles = {mole.name for mole in list(som1.molecules)}
    som2_moles = {mole.name for mole in list(som2.molecules)}
    reactions = self.get_edge_data(som1, som2)[cn.REACTION]
    # all reactions (in an edge), should create a single PathComponent
    nodes1 = []
    nodes2 = []
    reaction_labels = []
    for r in reactions:
      reaction = self.simple.getReaction(r)
      if reaction.category == cn.REACTION_n_1:
        sources = {r.molecule.name for r in reaction.reactants}
        destinations = {p.molecule.name for p in reaction.products}
      elif reaction.category == cn.REACTION_1_n:
        sources = {p.molecule.name for p in reaction.products}
        destinations = {r.molecule.name for r in reaction.reactants}
        # for any reaction that addes arcs, len(nodes2)==1
      node2 = list(destinations.intersection(som2_moles))[0]
      for node1 in list(sources.intersection(som1_moles)):
        nodes1.append(node1)
        nodes2.append(node2)
        reaction_labels.append(reaction.label)
    error_cycle.append(cn.PathComponents(node1=nodes1, 
                                        node2=nodes2,
                                        reactions=reaction_labels))
    self.type_two_errors.append(error_cycle)  

  def checkTypeTwoError(self):
    """
    Check Type II Error (cycles) of a MESGraph.
    If there is at least one cycle, 
    report an error message, related reactions
    and return True.
    If there is no cycle, return False. 
    :return bool:
    """
    graph = nx.DiGraph()
    graph.add_edges_from(self.edges)
    cycles = list(nx.simple_cycles(graph))
    if len(cycles) == 0:
      return False
    else:
      for cycle in cycles:
        self.addTypeTwoError(cycle)
        if not self.type_two_error:
          self.type_two_error = True
      return True

  def checkTypeFiveError(self):
    """
    Check Type V Error (cycles) of a MESGraph.
    If there is at least one cycle, 
    add cycle to self.type_five_errors.
    The biggest difference between type II error
    is that type five is for multi-multi reactions,
    so the cycle is reported by SOM-level. 
    :return bool:
    """
    graph = nx.DiGraph()
    graph.add_edges_from(self.edges)
    cycles = list(nx.simple_cycles(graph))
    if len(cycles) == 0:
      return False
    else:
      self.type_five_errors = cycles
      return True

  def analyze(self, reactions=None, error_details=True):
    """
    Sort list of reactions and process them.
    Add arcs or sending error messages using
    checkTypeOneError or checkTypeTwoError.
    :param list-Reaction reactions:
    :return str:
    """
    if reactions is None:
      reactions = self.simple.reactions
    # Associate the reaction category with the function
    # that processes that category
    report = NULL_STR
    reaction_dic = {
        cn.REACTION_1_1: self.processUniUniReaction,
        cn.REACTION_1_n: self.processUniMultiReaction,
        cn.REACTION_n_1: self.processMultiUniReaction,
        cn.REACTION_n_n: self.addMultiMultiReaction,
        }
    # Process each type of reaction
    for category in reaction_dic.keys():
      for reaction in [r for r in reactions if r.category == category]:
        func = reaction_dic[category]
        func(reaction)
    #
    self.checkTypeTwoError()    
    #
    if error_details:
      # if (len(self.type_one_errors)==0) and (len(self.type_two_errors)==0):
      #   report = report + "No error found in uni-uni and mulit-uni(uni-multi) reactions.\n"
      #
      for error_path in self.type_one_errors:
        path_report = self.printSOMPath(error_path.node1, error_path.node2)
        # print(path_report)
        report = report + "%s\n" % path_report
        # print("\nHowever, the following reaction(s)") 
        report = report + "However, the following reaction(s)\n"
        for arc_reaction in error_path.reactions:
          # print(self.simple.getReaction(arc_reaction).makeIdentifier(is_include_kinetics=False)) 
          report = "%s%s\n" % (report, self.simple.getReaction(arc_reaction).makeIdentifier(is_include_kinetics=False))
        # print("imply " + error_path.node1, cn.LESSTHAN,  error_path.node2)
        report = report + "imply %s %s %s\n" % (error_path.node1, cn.LESSTHAN, error_path.node2)
        # print("------------------------------------")
        report = report + "------------------------------------\n"
      #print("************************************")
      #
      # print("We Do have type II Errors", self.type_two_errors)
      for cycle in self.type_two_errors:
        for idx, path_comp in enumerate(cycle):
          nodes1 = collections.deque(path_comp.node1)
          nodes2 = collections.deque(path_comp.node2)
          if idx < len(cycle)-1:
            next_nodes1 = collections.deque(cycle[idx+1].node1)
          else:
            next_nodes1 = collections.deque(cycle[0].node1)
          reactions = collections.deque(path_comp.reactions)
          # print SOM path between node elements
          if len(nodes1)>1:
            for node_idx in range(len(nodes1)-1):
              path_report = self.printSOMPath(nodes1[node_idx], nodes1[node_idx+1])
              report = report + "%s\n" % path_report
          if not set(nodes2).intersection(set(next_nodes1)):
            path_report = self.printSOMPath(nodes2[0], next_nodes1[0])
            report = report + "%s\n" % path_report
         #
          while nodes1:
            report = report + "\n%s %s %s by reaction:\n" % (nodes1[0], cn.LESSTHAN, nodes2[0])
            arc_reaction = self.simple.getReaction(reactions[0])
            report = report + "%s\n" % (arc_reaction.makeIdentifier(is_include_kinetics=False))
            nodes1.popleft()
            nodes2.popleft()
            reactions.popleft()
        report = report + "%s" % ("------------------------------------")
      # report = report + "%s\n" % ("*************************************************************")
    # Process multi-multi reactions only if there's no elementary errors
    if len(self.type_one_errors)==0 and len(self.type_two_errors)==0:
      sub_multimulti = self.multimulti_reactions
      unsuccessful_load = 0
      max_loop = 0
      while (self.multimulti_reactions) and (max_loop<5):
        max_loop  = max_loop + 1
        flag_loop = [False] * len(self.multimulti_reactions)
        for idx, multimulti in enumerate(self.multimulti_reactions):
          result = self.processMultiMultiReaction(multimulti)
          if result is None:
            print("This reaction returned None") 
            print(multimulti)
            pass
          else:
            flag_loop[idx] = result
        # if nothing was processed, quit the while loop
        if sum(flag_loop)==0:
          break
        # if at least one was processed, subset unpressed reactions
        self.multimulti_reactions = [self.multimulti_reactions[idx] for idx, tr \
                                     in enumerate(flag_loop) if not tr]
      # check SOM cycles (type V error)
      self.checkTypeFiveError()
      # if len(self.type_three_errors)==0 and \
      #    len(self.type_four_errors)==0 and \
      #    len(self.type_five_errors)==0:
      #   # print("No error found in multi-multi reactions.")
      #   # print("*************************************************************")
      #   report = report + "%s\n%s\n" % (
      #       "No error found in multi-multi reactions.", 
      #       "*************************************************************")
      #
      # if error_details:
      #   if self.type_three_errors:
      #     print("We have type III errors\n", self.type_three_errors)
      #   else:
      #     print("We don't have type III errors")
      #   if self.type_four_errors:
      #     print("We have type IV errors\n", self.type_four_errors)
      #   else:
      #     print("We don't have type IV errors")
      #   if self.type_five_errors:
      #     print("We have type V errors\n", self.type_five_errors)
      #   else:
      #     print("We don't have type V errors")
    #
    self.identifier = self.makeId()
    print("Model analyzed...")
    if self.type_one_errors or \
        self.type_two_errors or \
        self.type_three_errors or \
        self.type_four_errors or \
        self.type_five_errors:
      print("At least one error found.")
    else:
      print("No error detected.")
    #return self
    return report



"""
Pythonic representation of an SBML model with some extensions.
- moietys (functional groups within a molecule)
- molecules (species)
- reactions
SimpleSBML extracts all information required from an SBML model to avoid
saving the libsbml object (since these objects are fragile with python
garbage collection).
"""

from SBMLLint.common import constants as cn
from SBMLLint.common.moiety import Moiety, MoietyStoichiometry
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.common import util

import collections
import os.path
import numpy as np
import sys
import libsbml
import urllib3
import warnings
import zipfile


TYPE_MODEL = "type_model"  # libsbml model
TYPE_XML = "type_xml"  # XML string
TYPE_ANTIMONY = "type_xml"  # Antimony string
TYPE_FILE = "type_file" # File reference

# filename: name of file processed
# number: index of item
# model: libsbml.Model
IteratorItem = collections.namedtuple('IteratorItem',
    'filename number model')


class SimpleSBML(object):
  """
  This class address stability of the underlying libsbml 
  (libsbml) library that seems not to survive garbage collection
  by python (e.g., returning a libsbml object to a caller.) As
  a result, no libsbml object is maintained by SimpleSBML instances.
  """

  def __init__(self):
    """
    Initializes instance variables
    """
    self.moietys = []
    self.molecules = []
    self.reactions = []

  def initialize(self, model_reference):
    """
    Initializes the instance variables in the model.
    :param str or libsbml.model: for str may be path or model string
       and file/str may be xml or antimony.
    """
    if util.isSBMLModel(model_reference):
      model = model_reference
    else:
      xml = util.getXML(model_reference)
      reader = libsbml.SBMLReader()
      document = reader.readSBMLFromString(xml)
      util.checkSBMLDocument(document, model_reference=model_reference)
      model = document.getModel()
    # Do the initializations
    self.reactions = self._getReactions(model)
    self.molecules = self._getMolecules()
    self.moietys = self._getMoietys()

  def _getReactions(self, model):
    reactions = []
    for nn in range(model.getNumReactions()):
      simple_reaction = Reaction(model.getReaction(nn))
      reactions.append(simple_reaction)
    return reactions

  def getReaction(self, label):
    """
    :param str label: label for the reaction
    :return Reaction/None:
    """
    reactions = [r for r in self.reactions if r.label == label]
    if len(reactions) > 1:
      raise ValueError("Two reactions with the same label: %s" %
          label)
    if len(reactions) == 0:
      return None
    return reactions[0]

  def _getMoietys(self):
    """
    Sees if there is a valid moiety structure.
    If not, the molecule is a single moiety.
    """
    moietys = []
    for molecule in self.molecules:
      try:
        new_moietys = ([m_s.moiety 
          for m_s in molecule.moiety_stoichiometrys])
      except ValueError:
        new_moietys = [Moiety(molecule.name)]
      moietys.extend(new_moietys)
    return util.uniqueify(moietys)

  def _getMolecules(self):
    """
    :return dict: key is species name, value is species object
    """
    molecules = []
    for reaction in self.reactions:
      molecules.extend(MoleculeStoichiometry.getMolecules(
          reaction.reactants))
      molecules.extend(MoleculeStoichiometry.getMolecules(
          reaction.products))
    return util.uniqueify(molecules)

  def getMolecule(self, name):
    """
    Finds and returns molecule with given name
    Return None if there is no such molecules
    :param str name:
    """
    molecules = [m for m in self.molecules if m.name == name]
    if len(molecules) > 1:
      raise ValueError("Duplicate names in simple.molecules.")
    elif len(molecules) == 1:
      return molecules[0]
    else:
      return None

  def add(self, element):
    """
    Adds an element of the type to its list
    """
    type_list = {
        Moiety: self.moietys,
        Molecule: self.molecules,
        Reaction: self.reactions,
        }
    this_list = type_list[element.__class__]
    appended_list = list(this_list)
    appended_list.append(element)
    new_list = util.uniqueify(appended_list)
    if len(new_list) > len(this_list):
      this_list.append(element)
    
  def remove(self, element):
    """
    Removes an element of the type to its list
    """
    type_list = {
        Moiety: self.moietys,
        Molecule: self.molecules,
        Reaction: self.reactions,
        }
    this_list = type_list[element.__class__]
    if element in this_list:
      this_list.remove(element)

###################### FUNCTIONS #############################
def readURL(url):
  """
  :param str url:
  :return str: file content
  """
  def do():
    http = urllib3.PoolManager()
    response = http.request('GET', url)
    return response.data.decode("utf-8") 
 # Catch bogus warnings
  with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    result = do()
  return result
  
def getZipfilePaths(data_dir=cn.BIOMODELS_DIR,
    zip_filename=cn.BIOMODELS_ZIP_FILENAME):
  """
  :param str data_dir: absolute path of the directory containing
      the xml files
  :param str zip_filename: name of the zipfile to process
  :return list-str, ZipFile: list of file paths, ZipFile object for file
  """
  path = os.path.join(data_dir, zip_filename)
  zipper = zipfile.ZipFile(path, "r")
  files = [f.filename for f in zipper.filelist]
  return files, zipper
  
def modelIterator(initial=0, final=1000, data_dir=cn.BIOMODELS_DIR,
    zip_filename=cn.BIOMODELS_ZIP_FILENAME):
  """
  Iterates across all models in a data directory.
  :param int initial: initial file to process
  :param int final: final file to process
  :param str data_dir: absolute path of the directory containing
      the xml files
  :param str zip_filename: name of the zipfile to process. If
      None, then looks for XML files in the directory.
  :return IteratorItem:
  """
  # Handle zip vs. XML file
  if zip_filename is not None:
    files, zipper = getZipfilePaths(
        data_dir=data_dir, zip_filename=zip_filename)
  else:
    files = [f for f in os.listdir(data_dir) if f[-4:] == ".xml"]
  # Functions for file types
  def readXML(filename):
    path = os.path.join(data_dir, filename)
    with open(path, 'r') as fd:
      lines = ''.join(fd.readlines())
    return lines
  def readZip(filename):
    with zipper.open(filename, 'r') as fid:
      lines = fid.read()
    return lines
  #
  if zip_filename is not None:
    read_func = readZip
  else:
    read_func = readXML
  begin_num = max(initial, 0)
  num = begin_num - 1
  end_num = min(len(files), final)
  for filename in files[begin_num:end_num]:
    num += 1
    lines = read_func(filename)
    if isinstance(lines, bytes):
      lines = lines.decode("utf-8") 
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromString(lines)
    model = document.getModel()
    iterator_item = IteratorItem(filename=filename,
        model=model, number=num)
    yield iterator_item



"""
Creates a full Antimony model from reaction strings.
This is done by:
1. Parsing the reactions for symbols used.
2. Creating assignment statements for symbols.
3. Append these assignments to the list of reactions.
Assumes that name segments are separated by a DOUBLEUNDERSCORE.

TODO:
1. Keep state of updates to model
2. findCandidateRenames - finds names that should be renamed
   to moiety stoichiometries (e.g. *[0-9]_. Returns a dict
   of current name (key) and proposed name change (value)
3. changeName - inputs a dict of old and new name. Changes names
   from longest to shortest.
"""

import pandas as pd
import sys

EXTRANEOUS = ["+", "->", "-", "(", ")", "*", "\n", ";", ","]
UNDERSCORE = "_"
DOUBLEUNDERSCORE = "__"
IGNORE_SYMBOLS = ["pow", "/"]

class ModelMaker(object):
  """
  Creates an antimony models
  """

  def __init__(self, in_arg):
    """
    :param list-str/str reaction_strs/path:
      list-str - list of reactions
      str - file path or string of reactions
    """
    self._reaction_strs = self._getReactionstrs(in_arg)
    self.symbols = None  # symbols found
    self.model_str = None  # full model

  def _getReactionstrs(self, in_arg):
    if isinstance(in_arg, list):
      return in_arg  # list of reactions
    else:
      if "->" in in_arg:
        return in_arg.split("\n")
      else:
        with open(in_arg, "r") as fd:
          return [l for l in fd.readlines()]

  @staticmethod
  def _splitNumber(stg):
    """
    Extracts an ending integer from a string if there is one.
    :return str, int: prefix, number
    """
    prefix = stg
    num = None
    for idx in range(len(stg)):
      try:
        _ = int(prefix[-1])
        if num is None:
          num = prefix[-1]
        else:
          num = prefix[-1] + num
        prefix = prefix[:-1]
      except:
        break
    if num is None:
      return prefix, None
    else:
      return prefix, int(num)

  @classmethod
  def _makeRepetitionNames(cls, symbol, exclude_funcs=None):
    """
    Changes a symbol name to use moiety-stoichiometry repetitions 
    if it has repetition numbers. Has arguments to selectively
    exclude symbols
    :param str symbol: candidate repetition name
    :param list-BooleanFunction exclude_funcs: returns True if name
        should be excluded for consideration as a repetition count.
    :return str:
    """
    if exclude_funcs is None:
      exclude_funcs = []
    splits = symbol.split(DOUBLEUNDERSCORE)
    is_changed = False
    for idx, stg in enumerate(splits):
      is_ignore = False
      for func in exclude_funcs:
        if func(stg):
          is_ignore = True
      if not is_ignore:
        moiety, stoich = cls._splitNumber(stg)
        if stoich is not None:
          is_changed = True
          splits[idx] = "%s%s%s" % (moiety, UNDERSCORE, stoich)
    if is_changed:
      return DOUBLEUNDERSCORE.join(splits)
    else:
      return symbol

  def getCandidateRenames(self, exclude_funcs=None):
    """
    Finds symbols that should be renamed to use repetition counts.
    :param list-BooleanFunction exclude_funcs: applied to name parts
    :return dict: key is current name; value is proposed name
    """
    if self.symbols is None:
      self.extractSymbols()
    rename_dict = {s: self.__class__._makeRepetitionNames(s, exclude_funcs=exclude_funcs)
        for s in self.symbols}
    rename_dict = {k: v for k, v in rename_dict.items() if k != v}
    return rename_dict

  def replaceSymbols(self, symbol_dict, is_sort=True):
    """
    Replaces the symbol (key) with its value.
    :param dict symbol_dict:
    :param bool is_sort: sort replacement strings by size to
        avoid substrint replacements
    Updates self.model_str
    :return str:
    """
    ser = pd.Series([len(k) for k in symbol_dict.keys()])
    ser.index = [s for s in symbol_dict.keys()]
    if is_sort:
      ser = ser.sort_values(ascending=False)
    for k in ser.index:
      self.model_str = self.model_str.replace(k, symbol_dict[k])
    return self.model_str

  def makeModelStr(self):
    """
    Creates creates a full antimony model.
    :return str antimony_model: string of antimony model
    """
    model_strs = list(self._reaction_strs)
    [model_strs.append("%s = 0" % s) for s in self.extractSymbols()]
    self.model_str = '\n'.join(model_strs)
    return self.model_str
 
  def makeModelFile(self, out_path):
    """
    Creates creates a full antimony model.
    :param str out_path: File that will be written with the model
    """
    modelstr = self.makeModelStr()
    with open(out_path, "w") as fd:
      fd.write(modelstr)

  def extractSymbols(self):
    """
    Extracts symbols from an antimony strings.
    :param list-str reaction_strs: List of reaction strings
    :return list-str: list of unique symbols found
    """
    self.symbols = []
    #
    for line in self._reaction_strs:
      # Remove the reaction label
      pieces = line.split(":")
      if len(pieces) == 2:
        piece = pieces[1]
      else:
        piece = pieces[0]
      for stg in EXTRANEOUS:
        piece = piece.replace(stg, " ")
      line_symbols = piece.split(" ")
      line_symbols = [s for s in line_symbols if len(s) > 0]
      for sym in line_symbols:
        try:
          float(sym)
        except:
          self.symbols.append(sym)
    # Get unique symbols and sort
    self.symbols = [s for s in set(self.symbols)
        if not s in IGNORE_SYMBOLS]
    self.symbols.sort()
    return self.symbols


if __name__ == '__main__':
  in_arg = sys.stdin.read()
  maker = ModelMaker(in_arg)
  sys.stdout.write(maker.makeModelStr())



"""Script for running mass balance checking tools."""


from SBMLLint.common import config
from SBMLLint.common import constants as cn
from SBMLLint.common.simple_sbml import SimpleSBML
from SBMLLint.common import util
from SBMLLint.games.games_pp import GAMES_PP
from SBMLLint.games.games_report import GAMESReport
from SBMLLint.moiety_analysis.moiety_comparator import MoietyComparator

import os
import sys
import libsbml

TYPE_I = "type1"
TYPE_II = "type2"
TYPE_III = "type3"
CANCELING = "canceling"
ECHELON = "echelon"
GAMES = "games"


def lint(model_reference=None, 
    file_out=sys.stdout,
    mass_balance_check=GAMES,
    config_fid=None,
    is_report=True,
    implicit_games=False):
  """
  Reports on errors found in a model
  :param str model_reference: 
      libsbml_model file in
      file, antimony string, xml string
  :param TextIOWrapper model_fid: fid for an XML file
  :param TextIOWrapper file_out:
  :param str mass_balance_check: how check for mass balance
  :param TextIOWrapper config_fid: readable stream
  :param bool is_report: print result
  :return MoietyComparatorResult/null/None:
  """
  config.setConfiguration(fid=config_fid)
  config_dct = config.getConfiguration()
  if util.isSBMLModel(model_reference):
    model = model_reference
  else:
    xml = util.getXML(model_reference)
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromString(xml)
    util.checkSBMLDocument(document)
    model = document.getModel()
  #
  simple = SimpleSBML()
  simple.initialize(model)
  if mass_balance_check==cn.MOIETY_ANALYSIS:
    result = MoietyComparator.analyzeReactions(simple)
    if is_report:
      for line in result.report.split('\n'):
          file_out.write("%s\n" % line)
    return result
  elif mass_balance_check == GAMES:
    if implicit_games:
      for ignored in config_dct[cn.CFG_IGNORED_MOLECULES]:
        simple = removeIgnored(simple, ignored)
    m = GAMES_PP(simple)
    games_result = m.analyze(simple.reactions)
    if games_result and is_report:
      gr = GAMESReport(m, explain_threshold=config_dct[cn.CFG_GAMES_THRESHOLD])
      errortype_dic = {TYPE_I: gr.reportTypeOneError,
                       TYPE_II: gr.reportTypeTwoError,
                       TYPE_III: gr.reportTypeThreeError,
                       CANCELING: gr.reportCancelingError,
                       ECHELON: gr.reportEchelonError
                      }
      for errors in m.error_summary:
        for category in errortype_dic.keys():
          if errors.type == category:
            func = errortype_dic[category]            
            report, _ = func(errors.errors, explain_details=True)
            print(report)
    return games_result
  else:
    print ("Specified method doesn't exist")
    return None

def removeIgnored(simple, ignored):
  """
  Remove an ignored molecule
  from all reactions in a simpleSBML model.
  :param SimpleSBML simple:
  :param str ignored: a molecule name
  :return SimpleSBML:
  """
  modified_reactions = []
  for r in simple.reactions:
    modified = False
    reactant_names = [reactant.molecule.name for reactant in r.reactants]
    product_names = [product.molecule.name for product in r.products]
    if ignored in reactant_names:
      r.reactants = [ms for ms in r.reactants 
          if ms.molecule.name != ignored]
      modified = True
    if ignored in product_names:
      r.products = [ms for ms in r.products
           if ms.molecule.name != ignored]
      modified = True
    r.identifier = r.makeIdentifier()
    r.category = r.getCategory()
    if modified:
      modified_reactions.append(r)
  return simple



from setuptools import setup, find_packages

INSTALL_REQUIRES = [
    "matplotlib",
    "networkx",
    "nose",
    "numpy",
    "pandas",
    "python-libsbml",
    "pyyaml",
    "scipy",
    "urllib3",
    "xlrd",
    ]

def doSetup(install_requires):
  setup(
      name='SBMLLint',
      version='1.0.11',
      author='Woosub Shin, Joseph L. Hellerstein',
      author_email='jlheller@uw.edu',
      packages=find_packages(exclude=['tests', 'analysis',
          'notebook', 'docs']),
      scripts=[
          'SBMLLint/tools/games',
          'SBMLLint/tools/games.bat',
          'SBMLLint/tools/moiety_analysis',
          'SBMLLint/tools/moiety_analysis.bat',
          'SBMLLint/tools/lp_analysis',
          'SBMLLint/tools/lp_analysis.bat',
          'SBMLLint/tools/make_moiety_structure',
          'SBMLLint/tools/make_moiety_structure.bat',
          'SBMLLint/tools/print_reactions',
          'SBMLLint/tools/print_reactions.bat',
          ],
      url='https://github.com/ModelEngineering/SBMLLint',
      description='Linter for SBML models.',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      package_dir={'SBMLLint': 'SBMLLint'},
      install_requires=install_requires,
      include_package_data=True,
      data_files=[('data/biomodels',
          ['data/biomodels/biomodels.zip'])],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',      # Define that your audience are developers
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'License :: OSI Approved :: MIT License',   # Again, pick a license
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
        ],
      )


if __name__ == '__main__':
  doSetup(INSTALL_REQUIRES)



"""Stoichiometry Matrix."""

from SBMLLint.common import constants as cn
from SBMLLint.common.molecule import Molecule, MoleculeStoichiometry
from SBMLLint.common.reaction import Reaction
from SBMLLint.games.som import SOM
from SBMLLint.common.simple_sbml import SimpleSBML

import numpy as np
import pandas as pd
from scipy.optimize import linprog
import warnings


class StoichiometryMatrix(object):
  """
  Creates a full stoichiometry matrix from simpleSBML 
  that correctly incorporates boundary species 
  and use linear programming to determine stoichiometric consistency.
  """    
  def __init__(self, simple=None):
    self.reactions = self._getNonBoundaryReactions(simple)
    self.molecules = self._getNonBoundaryMolecules(simple)
    self.stoichiometry_matrix = self.makeStoichiometryMatrix()
    self.consistent = None
    self.result = None

  def _getNonBoundaryReactions(self, simple):
    """
    Get list of non-boundary reacetions
    :param SimpleSBML simple:
    :return list-Reaction:
    """
    reactions = []
    for reaction in simple.reactions:
      if reaction.category != cn.REACTION_BOUNDARY:
        reactions.append(reaction)
    return reactions

  def _getNonBoundaryMolecules(self, simple):
    """
    Get list of non-boundary molecules
    :param SimpleSBML simple:
    :return list-Molecule.name:
    """
    molecules = set()
    for reaction in self.reactions:
      reactants = {r.molecule.name for r in reaction.reactants}
      products = {r.molecule.name for r in reaction.products}
      molecules = molecules.union(reactants)
      molecules = molecules.union(products)
    return list(molecules)

  def makeStoichiometryMatrix(self):
    """
    Creates a full stoichiometry matrix
    using non-boundary reactions.
    Helped by https://gist.github.com/lukauskas/d1e30bdccc5b801d341d
    :return pd.DataFrame:
    """
    reaction_labels = [r.label for r in self.reactions]
    stoichiometry_matrix = pd.DataFrame(0.0, index=self.molecules, columns=reaction_labels)
    for reaction in self.reactions:
      reactants = {r.molecule.name:r.stoichiometry for r in reaction.reactants}
      products = {p.molecule.name:p.stoichiometry for p in reaction.products}
      reaction_molecules = list(set(reactants.keys()).union(products.keys()))
      for molecule_name in reaction_molecules:
        net_stoichiometry = products.get(molecule_name, 0.0) - reactants.get(molecule_name, 0.0)
        stoichiometry_matrix[reaction.label][molecule_name] = net_stoichiometry
    return stoichiometry_matrix

  def isConsistent(self, is_report_warning=True):
    """
    Runs linear programmming (LP) to determine 
    stoichiometric inconsistency. 
    If consistent return True,
    else return False. 
    :param bool is_report_warning: report optimization warnings
    :return bool:
    """
    s_matrix_t = self.stoichiometry_matrix.T
    # number of molecules
    nmet = s_matrix_t.shape[0]
    # number of reactions
    nreac = s_matrix_t.shape[1]
    #
    b = np.zeros(nmet)
    c = np.ones(nreac)
    # Linear programming. c is constraint (here, zero), 
    # b is vector of possible values for molecule vector. 
    if not is_report_warning:
      warnings.simplefilter("ignore")
    try:
      res = linprog(c, A_eq=s_matrix_t, b_eq=b, bounds=(1, None))
      self.result = res
      is_success = True
    except:
      is_success = False
    if not is_success:
      msg = "*** Failed to solve the stoichiometry matrix."
      raise RuntimeError(msg)
    if res.status == 0:
      self.consistent = True
    else:
      self.consistent = False
    #
    return self.consistent



















"""Ensures that the correct root path is present."""
import os
import sys

def addPath():
  root_dir = sys.path[0]
  for _ in range(2):
    root_dir, _ = os.path.split(root_dir)
  sys.path.insert(0, root_dir)

addPath()



"""Wraps methods that use Tellurium runs them in another process."""

from SBMLLint.common import exceptions

import argparse
import subprocess
import sys


def getInstalledPackages():
  """
  Returns list of installed packages.
  :return list-str"
  """
  reqs = subprocess.check_output([sys.executable,
      '-m', 'pip', 'freeze'])
  installed_packages = [r.decode().split('==')[0]
      for r in reqs.split()]
  return installed_packages


class TelluriumSandbox(object):
  """
  Runs python code in a separate process (sandbox).
  Note that there are two instances of this class that are
  created; one in the parent process (which uses the main and run
  methods) and one in the child (which uses the execute method).
  Usage:
    sandbox.run(method_name, string_argument)
  Outputs
    sandbox.return_code - should be 0
    sandbox.output - string result
  """

  def __init__(self, dependencies=["tellurium"]):
    """
    :param list-str dependencies: dependent packages
    """
    pkgs = getInstalledPackages()
    missing_pkgs = set(dependencies).difference(pkgs)
    if len(missing_pkgs) > 0:
      pkgs_stg = ", ".join(missing_pkgs)
      msg = "***Operation aborted. %s not installed." % pkgs_stg
      raise exceptions.MissingTelluriumError(msg)
    self.return_code = None
    self.output = None

  def echo(self, input_string):
    """
    Executes the sandboxed python codes in the child process.
    :param str input_string:
    :return str:
    Notes:
      1. Writes double newlines ("\n")
    """
    sys.stdout.writelines(input_string)

  @staticmethod
  def _convert(input_string):
    """
    Converts from io_wrapper to string
    """
    return ''.join([l for l in input_string])

  def getSBMLFromAntimony(self, input_string):
    """
    This method runs in a different process from the caller
    and so can import tellurium.
    """
    import tellurium as te
    # Convert input to string.
    inputs = self.__class__._convert(input_string)
    # Try different ways to load the model
    try:
      rr = te.loada(inputs)
    except:
      rr = te.loadSBMLModel(inputs)
    sbml = rr.getSBML()
    sys.stdout.writelines(sbml)

  def run(self, method, input_string):
    """
    Runs the child process from the parent process.
    Strips double newlines.
    """
    process = subprocess.run(['python', __file__, method],
        stdout=subprocess.PIPE,
        input=input_string, encoding='ascii')
    self.return_code = process.returncode
    self.output = process.stdout

  def main(self):
    """
    The first argument is the method to call.
    """
    cmd = "self.%s(sys.stdin)" % sys.argv[1]
    exec(cmd)


if __name__ == '__main__':
  sandbox = TelluriumSandbox()
  sandbox.main()
